/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.work.queue;

import com.google.common.collect.ImmutableMap;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.AsyncWorkQueueCustomExecutor;
import org.artifactory.api.repo.AsyncWorkQueueProviderService;
import org.artifactory.api.repo.WorkItem;
import org.artifactory.api.repo.WorkQueue;
import org.artifactory.common.ConstantValues;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.mbean.MBeanRegistrationService;
import org.artifactory.schedule.CachedThreadPoolTaskExecutor;
import org.artifactory.schedule.JobCommand;
import org.artifactory.schedule.JobName;
import org.artifactory.schedule.QuartzCommand;
import org.artifactory.schedule.Task;
import org.artifactory.schedule.TaskService;
import org.artifactory.schedule.TaskUser;
import org.artifactory.schedule.TaskUtils;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.util.CollectionUtils;
import org.artifactory.work.queue.mbean.WorkQueueMBean;
import org.artifactory.work.queue.mbean.airgap.AirGapWorkQueue;
import org.artifactory.work.queue.mbean.alpine.AlpineVirtualWorkQueue;
import org.artifactory.work.queue.mbean.alpine.AlpineWorkQueue;
import org.artifactory.work.queue.mbean.buildEvent.BuildRetentionWorkQueue;
import org.artifactory.work.queue.mbean.buildinfo.BuildInfoMigrationCalculationJobWorkQueue;
import org.artifactory.work.queue.mbean.cargo.CargoWorkQueue;
import org.artifactory.work.queue.mbean.checksum.ChecksumCalculationJobWorkQueue;
import org.artifactory.work.queue.mbean.chef.ChefMetadataWorkQueue;
import org.artifactory.work.queue.mbean.composer.ComposerExtractorWorkQueue;
import org.artifactory.work.queue.mbean.composer.ComposerIndexerWorkQueue;
import org.artifactory.work.queue.mbean.conan.ConanV2MigrationCalculationJobWorkQueue;
import org.artifactory.work.queue.mbean.conan.ConanWorkQueue;
import org.artifactory.work.queue.mbean.conda.CondaWorkQueue;
import org.artifactory.work.queue.mbean.cran.CranWorkQueue;
import org.artifactory.work.queue.mbean.debian.DebianCacheCoordinatesWorkQueue;
import org.artifactory.work.queue.mbean.debian.DebianVirtualWorkQueue;
import org.artifactory.work.queue.mbean.debian.DebianWorkQueue;
import org.artifactory.work.queue.mbean.docker.DockerWorkQueue;
import org.artifactory.work.queue.mbean.gems.GemsVirtualWorkQueue;
import org.artifactory.work.queue.mbean.gems.GemsWorkQueue;
import org.artifactory.work.queue.mbean.git.GitWorkQueue;
import org.artifactory.work.queue.mbean.ha.HaMessageWorkQueue;
import org.artifactory.work.queue.mbean.helm.HelmVirtualWorkQueue;
import org.artifactory.work.queue.mbean.helm.HelmWorkQueue;
import org.artifactory.work.queue.mbean.maven.MavenMetadataWorkQueue;
import org.artifactory.work.queue.mbean.maven.MavenPluginMetadataWorkQueue;
import org.artifactory.work.queue.mbean.mirror.MirrorFullSyncWorkQueue;
import org.artifactory.work.queue.mbean.mirror.MirrorHeartbeatWorkQueue;
import org.artifactory.work.queue.mbean.npm.conda.NpmReindexWorkQueue;
import org.artifactory.work.queue.mbean.npm.conda.NpmWorkQueue;
import org.artifactory.work.queue.mbean.nuget.NugetWorkQueue;
import org.artifactory.work.queue.mbean.opkg.OpkgWorkQueue;
import org.artifactory.work.queue.mbean.persistent.queue.PersistentQueueWorkQueue;
import org.artifactory.work.queue.mbean.pub.PubWorkQueue;
import org.artifactory.work.queue.mbean.puppet.PuppetRepoWorkQueue;
import org.artifactory.work.queue.mbean.puppet.PuppetWorkQueue;
import org.artifactory.work.queue.mbean.replication.ReplicationEventQueue;
import org.artifactory.work.queue.mbean.repo.path.checksum.RepoPathChecksumCalculationJobWorkQueue;
import org.artifactory.work.queue.mbean.storage.AsyncDeleteWorkQueue;
import org.artifactory.work.queue.mbean.swift.SwiftWorkQueue;
import org.artifactory.work.queue.mbean.terraform.TerraformWorkQueue;
import org.artifactory.work.queue.mbean.yum.RpmWorkQueue;
import org.artifactory.work.queue.mbean.yum.YumVirtualWorkQueue;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.concurrency.JvmConflictsGuard;
import org.quartz.JobDataMap;
import org.quartz.JobDetail;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.scheduling.concurrent.ConcurrentTaskExecutor;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import javax.annotation.PreDestroy;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;

/**
 * @author gidis
 */
@Service
public class AsyncWorkQueueServiceImpl implements AsyncWorkQueueProviderService, ContextCreationListener {
    public static final String JOB_QUEUE_PARAM = "queueService";
    private static final Logger log = LoggerFactory.getLogger(AsyncWorkQueueServiceImpl.class);
    private final ConcurrentHashMap<String, WorkQueue<WorkItem>> workQueueMap = new ConcurrentHashMap<>();
    @Autowired
    private TaskService taskService;

    @Override
    public void onContextCreated() {
        if (ConstantValues.workQueueJobEnabled.getBoolean()) {
            Task workQueueDoJobTask = TaskUtils.createRepeatingTask(WorkQueueJob.class,
                    TimeUnit.SECONDS.toMillis(ConstantValues.workQueueDoJobIntervalSecs.getLong()),
                    TimeUnit.SECONDS.toMillis(ConstantValues.workQueueDoJobIntervalSecs.getLong()));
            workQueueDoJobTask.addAttribute(JOB_QUEUE_PARAM, this);
            taskService.startTask(workQueueDoJobTask, false);
        }
    }

    @PreDestroy
    public void destroy() {
        taskService.cancelTasks(WorkQueueJob.class, true);
    }

    @Override
    public void closeAllQueues() {
        for (Map.Entry<String, WorkQueue<WorkItem>> queueEntry : workQueueMap.entrySet()) {
            WorkQueue<WorkItem> workQueue = queueEntry.getValue();
            try {
                workQueue.stopQueue();
            } catch (Exception e) {
                log.error("Failed to close queue " + queueEntry.getKey(), e);
            }
        }
    }

    @Override
    public int getEstimatedPendingTasksSize(String workQueueCallbackName) {
        WorkQueue<WorkItem> workQueue = workQueueMap.get(workQueueCallbackName);
        if (workQueue != null) {
            return workQueue.getQueueSize();
        }
        throw new RuntimeException("No work queue matches callback '" + workQueueCallbackName + "'");
    }

    @Override
    public WorkQueue<WorkItem> getWorkQueue(Method workQueueCallback, Object target) {
        String name = workQueueCallback.getName();
        String displayNameSuffix = StringUtils.EMPTY;

        if (CustomExecutor.get() != null) {
            if (StringUtils.isNotBlank(CustomExecutor.get().getNameSuffix())) {
                //append the custom executor suffix to the name, in order to get a unique name for the custom
                name += "-" + CustomExecutor.get().getNameSuffix();
                displayNameSuffix = " " + CustomExecutor.get().getNameSuffix();
            }

            if (StringUtils.isNotBlank(CustomExecutor.get().getRepoKey())) {
                String repoKey = CustomExecutor.get().getRepoKey();
                name += "-" + repoKey;
                displayNameSuffix += " " + repoKey;
            }
        }

        WorkQueue<WorkItem> workQueue = workQueueMap.get(name);
        if (workQueue == null) {
            synchronized (workQueueMap) {
                workQueue = workQueueMap.get(name);
                if (workQueue == null) {
                    try {
                        WorkQueueInfo info = getWorkerInfo(workQueueCallback, displayNameSuffix);
                        // if queue already exists for another method (under different key in the 'workQueueMap' map),
                        // return the existing queue and add it to the map for the current method name as well
                        WorkQueue<WorkItem> existingQueue = searchQueueUnderDifferentKey(info);
                        if (existingQueue != null) {
                            workQueue = existingQueue;
                        } else {
                            // the workQueue is limited to work per class. I think that we should pass the target
                            //class like we do with the method as a param to the workQueue and not cache it in the
                            //workQueue object attributes
                            workQueue =
                                    new WorkQueueImpl<>(info.getDisplayName(),
                                            getMaxNumberOfThreads(CustomExecutor.get(), info), target,
                                            CustomExecutor.get(), info.getConflictsGuard());
                            WorkQueueMBean mBean = createMBean(workQueue, info);
                            ContextHelper.get().beanForType(MBeanRegistrationService.class)
                                    .register(mBean, "Work Queue", info.displayName);
                        }
                        workQueueMap.put(name, workQueue);
                    } catch (Exception e) {
                        log.error("Failed to initialize work queue {}", name, e);
                    }
                }
            }
        }
        return workQueue;
    }

    private int getMaxNumberOfThreads(AsyncWorkQueueCustomExecutor customExecutor, WorkQueueInfo info) {
        int maxThreads = info.getMaxNumberOfThreads();
        if (customExecutor != null && customExecutor.getThreadsCount() != null &&
                customExecutor.getThreadsCount().get() != null) {
            maxThreads = customExecutor.getThreadsCount().get();
        }
        return maxThreads;
    }

    /**
     * Search for queue with the same name as the info object under any kind of key in the workQueueMap map.
     *
     * @return WorkQueue if found, null otherwise
     */
    private WorkQueue<WorkItem> searchQueueUnderDifferentKey(WorkQueueInfo info) {
        Collection<WorkQueue<WorkItem>> queues = workQueueMap.values();
        if (CollectionUtils.isNullOrEmpty(queues)) {
            return null;
        }
        return queues.stream()
                .filter(queue -> queue.getName().equals(info.getDisplayName()))
                .findFirst()
                .orElse(null);
    }

    @Override
    public Map<String, WorkQueue<WorkItem>> getExistingWorkQueues() {
        return workQueueMap;
    }

    private WorkQueueMBean createMBean(WorkQueue<WorkItem> workQueue, WorkQueueInfo info)
            throws InvocationTargetException, NoSuchMethodException, InstantiationException, IllegalAccessException {
        try {
            Constructor<? extends WorkQueueMBean> constructor =
                    info.getMBeanClass().getConstructor(WorkQueueMBean.class);
            //noinspection JavaReflectionInvocation
            return constructor.newInstance(workQueue);
        } catch (Exception e) {
            log.error("Failed to create mbean for work queue {}", workQueue.getName());
            throw e;
        }
    }

    private WorkQueueInfo getWorkerInfo(Method workQueueCallback, String displayNameSuffix) {
        String name = workQueueCallback.getName();
        String className = workQueueCallback.getDeclaringClass().getName();
        switch (name) {
            case "calculateMavenMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.mvnMetadataCalculationWorkers.getInt();
                String displayName = "Maven Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, MavenMetadataWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateMavenPluginsMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.mvnMetadataPluginCalculationWorkers.getInt();
                String displayName = "Plugin Maven Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, MavenPluginMetadataWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateYumVirtualMetadataAsync":
            case "calculateYumVirtualMetadata": {
                int maxNumberOfThreads = ConstantValues.yumVirtualMetadataCalculationWorkers.getInt();
                String displayName = "Yum Virtual Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, YumVirtualWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateRpmMetadataAsync":
            case "calculateRpmMetadataSync": {
                int maxNumberOfThreads = ConstantValues.rpmMetadataCalculationWorkers.getInt();
                String displayName = "Rpm Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, RpmWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateHelmMetadataAsync":
            case "calculateHelmMetadataSync": {
                int maxNumberOfThreads = ConstantValues.helmMetadataCalculationWorkers.getInt();
                String displayName = "Helm Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, HelmWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateHelmVirtualMetadataSync": {
                int maxNumberOfThreads = ConstantValues.helmVirtualMetadataCalculationWorkers.getInt();
                String displayName = "Helm Virtual Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, HelmVirtualWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateAlpineMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.alpineMetadataCalculationWorkers.getInt();
                String displayName = "Alpine Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, AlpineWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateAlpineVirtualMetadataSync": {
                int maxNumberOfThreads = ConstantValues.alpineVirtualMetadataCalculationWorkers.getInt();
                String displayName = "Alpine Virtual Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, AlpineVirtualWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculatePubMetadataAsync":
            case "calculatePubMetadataSync": {
                int maxNumberOfThreads = ConstantValues.pubMetadataCalculationWorkers.getInt();
                String displayName = "Pub Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, PubWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateSwiftMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.swiftMetadataCalculationWorkers.getInt();
                String displayName = "Swift Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, SwiftWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateCargoMetadataAsync":
            case "calculateCargoMetadataSync": {
                int maxNumberOfThreads = ConstantValues.cargoMetadataCalculationWorkers.getInt();
                String displayName = "Cargo Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, CargoWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateConanMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.conanMetadataCalculationWorkers.getInt();
                String displayName = "Conan Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, ConanWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateCranMetadataSync":
            case "calculateCranMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.cranMetadataCalculationWorkers.getInt();
                String displayName = "CRAN Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, CranWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateCranVirtualMetadataSync": {
                int maxNumberOfThreads = ConstantValues.cranMetadataCalculationWorkers.getInt();
                String displayName = "CRAN virtual Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, CranWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateCondaMetadataSync":
            case "calculateCondaMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.condaMetadataCalculationWorkers.getInt();
                String displayName = "Conda Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, CondaWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "updateGitRepo":
            case "updateGitLocalRepo": {
                int maxNumberOfThreads = ConstantValues.gitRepositoriesWorkers.getInt();
                String displayName = "Git Repository Update" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, GitWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateNpmMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.npmMetadataCalculationWorkers.getInt();
                String displayName = "Npm Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, NpmWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "reindexNpmMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.npmReindexMetadataCalculationWorkers.getInt();
                String displayName = "Npm Reindex Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, NpmReindexWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "handlePackageDeployment":
            case "handlePackageDeletion": {
                int maxNumberOfThreads = ConstantValues.composerMetadataExtractorWorkers.getInt();
                String displayName = "Composer Metadata Extraction" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, ComposerExtractorWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "extractAndIndexAllComposerPackages":
            case "indexComposerPackageAndRepo": {
                int maxNumberOfThreads = ConstantValues.composerMetadataIndexWorkers.getInt();
                String displayName = "Composer Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, ComposerIndexerWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateCachedDebianCoordinates": {
                int maxNumberOfThreads = ConstantValues.debianCoordinatesCalculationWorkers.getInt();
                String displayName = "Debian Coordinates Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, DebianCacheCoordinatesWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateDebianMetadataInternalAsync":
            case "calculateDebianMetadataInternalSync": {
                int maxNumberOfThreads = ConstantValues.debianMetadataCalculationWorkers.getInt();
                String displayName = "Debian Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, DebianWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateOpkgMetadataInternalAsync":
            case "calculateOpkgMetadataInternalSync": {
                int maxNumberOfThreads = ConstantValues.opkgMetadataCalculationWorkers.getInt();
                String displayName = "Opkg Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, OpkgWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateDebianVirtualMetadataAsync":
            case "calculateAllDebianVirtualMetadataSync":
            case "calculateAllDebianVirtualMetadataAsync":
            case "calculateDebianVirtualMetadataSync": {
                int maxNumberOfThreads = ConstantValues.debianVirtualMetadataCalculationWorkers.getInt();
                String displayName = "Debian Virtual Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, DebianVirtualWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculatePuppetMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.puppetMetadataCalculationWorkers.getInt();
                String displayName = "Puppet Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, PuppetWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculatePuppetRepoMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.puppetRepoMetadataCalculationWorkers.getInt();
                String displayName = "Puppet Repository Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, PuppetRepoWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "addEtagForGems":
            case "updateCompactIndex": {
                int maxNumberOfThreads = ConstantValues.gemsCompactIndexWorkers.getInt();
                String displayName = "Gems Repository Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, GemsWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "handleIndexAddAfterCommit":
            case "handleAddAfterCommit": {
                int maxNumberOfThreads = ConstantValues.gemsAfterCommitWorkers.getInt();
                String displayName = "Gems After Commit" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, GemsWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "calculateVersionsIncrementally":
            case "calculateVersionsFull": {
                int maxNumberOfThreads = ConstantValues.gemsCompactIndexWorkers.getInt();
                String displayName = "Gems Virtual Repository Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, GemsVirtualWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "extractAndIndexChefCookbooks":
            case "extractAndIndexSingleChefCookbook":
            case "calculateVirtualRepoMetadata": {
                int maxNumberOfThreads = ConstantValues.chefMetadataIndexWorkers.getInt();
                String displayName = "Chef Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, ChefMetadataWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "offerRemoteReplicationEventInternal":
            case "putInboundReplicationEventsInternal": {
                int maxNumberOfThreads = ConstantValues.eventBasedReplicationWorkers.getInt();
                String displayName = "Event Based Replication" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, ReplicationEventQueue.class,
                        conflictsGuard(displayName));
            }
            case "deleteBuildAsync": {
                int maxNumberOfThreads = ConstantValues.buildRetentionWorkers.getInt();
                String displayName = "Build Retention Job" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, BuildRetentionWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "updateSha2": {
                String displayName = "SHA256 Migration" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.sha2MigrationJobQueueWorkers.getInt(), displayName,
                        ChecksumCalculationJobWorkQueue.class, conflictsGuard(displayName));
            }
            case "migrateBuildInfo": {
                String displayName = "Build Info Migration" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.buildInfoMigrationJobQueueWorkers.getInt(),
                        displayName, BuildInfoMigrationCalculationJobWorkQueue.class, conflictsGuard(displayName));
            }
            case "migrateConanToV2": {
                String displayName = "Conan V2 Migration" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.conanV2MigrationJobQueueWorkers.getInt(),
                        displayName, ConanV2MigrationCalculationJobWorkQueue.class, conflictsGuard(displayName));
            }
            case "updateRepoPathChecksum": {
                String displayName = "Path Checksum Migration" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.pathChecksumMigrationJobQueueWorkers.getInt(),
                        displayName, RepoPathChecksumCalculationJobWorkQueue.class, conflictsGuard(displayName));
            }
            case "notifyAsync": {
                String displayName = "HA event messages" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.haMessagesWorkers.getInt(), displayName,
                        HaMessageWorkQueue.class, conflictsGuard(displayName));
            }
            case "runAsyncAirGapReleaseBundleExport": {
                String displayName = "AirGapExport" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.airGapWorkers.getInt(), displayName, AirGapWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "processArtifactsAsync": {
                String displayName = "AirGapImport" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.airGapWorkers.getInt(), displayName, AirGapWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "deletePathAsync": {
                String displayName = "AsyncDeletePathItem" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.asyncDeleteWorkers.getInt(),
                        displayName, AsyncDeleteWorkQueue.class, conflictsGuard(displayName));
            }
            case "executeSyncCatalogIndex":
            case "executeAsyncCatalogIndex": {
                String displayName = "Docker Catalog Index" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.dockerIndexWorkers.getInt(), displayName, DockerWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "triggerQueue": {
                String displayName = "Persistent Queue triggering" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.persistentQueueTriggerNumberOfThreads.getInt(),
                        displayName, PersistentQueueWorkQueue.class, conflictsGuard(displayName));
            }
            case "createFullMetadataFeedSync": {
                String displayName = "Nuget Feed Calculation" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.nuGetV2FeedCalculationWorkers.getInt(), displayName,
                        NugetWorkQueue.class, conflictsGuard(displayName));
            }
            case "triggerAsyncHeartbeat": {
                String displayName = "Mirror Heartbeat" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.mirrorHeartbeatWorkers.getInt(),
                        displayName, MirrorHeartbeatWorkQueue.class,
                        new JvmConflictsGuard<>(ConstantValues.mirrorHeartbeatLockLeaseTimeMin.getLong()));
            }
            case "calculateTerraformMetadataAsync": {
                int maxNumberOfThreads = ConstantValues.terraformMetadataCalculationWorkers.getInt();
                String displayName = "Terraform Metadata" + displayNameSuffix;
                return new WorkQueueInfo(maxNumberOfThreads, displayName, TerraformWorkQueue.class,
                        conflictsGuard(displayName));
            }
            case "runFullSyncAsync":
            case "triggerFullSync": {
                String displayName = "Full sync triggering" + displayNameSuffix;
                return new WorkQueueInfo(ConstantValues.mirrorFullSyncWorkers.getInt(), displayName,
                        MirrorFullSyncWorkQueue.class,
                        conflictsGuard(NonStandardLockCategory.FED_REPO_FULL_SYNC.getName()));
            }
            default: {
                throw new RuntimeException(
                        "Unsupported work queue: the work queue: class" + className + " method: " + name +
                                " is not supported by this service");
            }
        }
    }

    private ConflictsGuard<String> conflictsGuard(String displayName) {
        return ContextHelper.get().beanForType(AddonsManager.class).addonByType(HaAddon.class)
                .getConflictsGuard(displayName);
    }

    // Job that scan work queues that has pending "stuck" tasks with no thread to execute them, will trigger thread
    //to execute the "stuck" task
    @JobCommand(jobName = JobName.WORK_QUEUE, description = "Invoke work queue tasks", singleton = true,
            clusterSingleton = false, allowedAffinity = false, schedulerUser = TaskUser.SYSTEM, manualUser =
            TaskUser.SYSTEM)
    public static class WorkQueueJob extends QuartzCommand {

        @Override
        protected void onExecute(JobExecutionContext callbackContext) throws JobExecutionException {
            try {
                //clear work queues of deleted fed repo
                clear(callbackContext);
                Map<String, WorkQueue<WorkItem>> queues = getQueues(callbackContext);
                log.trace("Found map with {} queues", queues.size());
                invokeIfNeeded(queues);
            } catch (Exception e) {
                log.error("An error occurred while running {} job. {}", this.getClass().getName(), e.getMessage());
                log.debug("An error occurred while running job.", e);
            }
        }

        private void clear(JobExecutionContext callbackContext) {
            CentralConfigDescriptor descriptor = ContextHelper.get().beanForType(CentralConfigService.class)
                    .getDescriptor();
            //remove work queue of custom executor if repo does not exist
            Set<Map.Entry<String, WorkQueue<WorkItem>>> entries = getQueues(callbackContext).entrySet();
            try {
                entries.removeIf(
                        entry -> {
                            if (entry.getValue().customExecutor() != null &&
                                    entry.getValue().customExecutor().getRepoKey() != null) {
                                boolean remove = !descriptor.getFederatedRepositoriesMap()
                                        .containsKey(entry.getValue().customExecutor().getRepoKey());
                                if (remove) {
                                    MBeanRegistrationService.unregister(
                                            MBeanRegistrationService.createObjectName("Work Queue",
                                                    entry.getValue().getName()));
                                    log.info("Removing WorkQueue {}", entry.getValue().getName());
                                }
                                return remove;
                            }
                            return false;
                        });
            } catch (UnsupportedOperationException e) {
                //Ignore on UnsupportedOperationException, in some cases getQueues() return ImmutableMap
            }
        }

        private Map<String, WorkQueue<WorkItem>> getQueues(JobExecutionContext callbackContext) {
            Map<String, WorkQueue<WorkItem>> queues = null;
            log.debug("Searching for job details");
            JobDetail jobDetail = callbackContext.getJobDetail();
            if (jobDetail != null) {
                JobDataMap jobDataMap = jobDetail.getJobDataMap();
                if (jobDataMap != null) {
                    log.debug("Resolving queues from the queue service");
                    AsyncWorkQueueProviderService queueService = (AsyncWorkQueueProviderService) jobDataMap
                            .get(JOB_QUEUE_PARAM);
                    queues = queueService.getExistingWorkQueues();
                }
            }
            return queues != null ? queues : ImmutableMap.of();
        }

        private void invokeIfNeeded(Map<String, WorkQueue<WorkItem>> queues) {
            if (queues != null && !queues.isEmpty()) {
                queues.values().forEach(queue -> {
                    //queue.availablePermits should never be bigher than getMaxNumberOfWorkers, unless if there is a
                    //bug that we do release() more than we need. this is why >= and not ==
                    if (queue.getQueueSize() > 0) {
                        //It has been decided to log on info. Re-consider in the future to convert to debug log
                        log.info("Invoking tasks on {}", queue.getName());

                        ConcurrentTaskExecutor executor =
                                queue.customExecutor() != null ? queue.customExecutor().getExecutor() : null;
                        if (executor == null) {
                            executor =
                                    ContextHelper.get().beanForType(CachedThreadPoolTaskExecutor.class);
                        }
                        queue.execute(executor);
                    }
                });
            }
        }
    }

    private static class WorkQueueInfo {
        private final int maxNumberOfThreads;
        private final String displayName;
        private final Class<? extends WorkQueueMBean> mBeanClass;
        private final ConflictsGuard<String> conflictsGuard;

        public WorkQueueInfo(int maxNumberOfThreads, String displayName, Class<? extends WorkQueueMBean> mBeanClass,
                @Nonnull ConflictsGuard<String> conflictsGuard) {
            this.maxNumberOfThreads = maxNumberOfThreads;
            this.displayName = displayName;
            this.mBeanClass = mBeanClass;
            this.conflictsGuard = conflictsGuard;
        }

        public int getMaxNumberOfThreads() {
            return maxNumberOfThreads;
        }

        public String getDisplayName() {
            return displayName;
        }

        public Class<? extends WorkQueueMBean> getMBeanClass() {
            return mBeanClass;
        }

        public ConflictsGuard<String> getConflictsGuard() {
            return conflictsGuard;
        }
    }
}
