/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.work.queue;

import com.google.common.annotations.VisibleForTesting;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.AsyncWorkQueueCustomExecutor;
import org.artifactory.api.repo.WorkItem;
import org.artifactory.api.repo.WorkQueue;
import org.artifactory.common.ConstantValues;
import org.artifactory.schedule.ArtifactoryConcurrentExecutor;
import org.artifactory.schedule.CachedThreadPoolTaskExecutor;
import org.artifactory.work.queue.mbean.WorkQueueMBean;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.scheduling.concurrent.ConcurrentTaskExecutor;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.time.Duration;
import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.Semaphore;

/**
 * A work queue that eliminates jobs based on a passed criteria.<p>
 * The thread offering work for this queue might start working on the queue, and may work on all the queued items.
 *
 * @author Yossi Shaul
 */
public class WorkQueueImpl<T extends WorkItem> implements WorkQueue<T>, WorkQueueMBean {
    private static final Logger log = LoggerFactory.getLogger(WorkQueueImpl.class);

    private final NonBlockingOnWriteQueue<T> queue;
    private final String name;
    private volatile int workers;
    private final Semaphore semaphore;
    private final Object target;

    @Nullable
    private AsyncWorkQueueCustomExecutor customExecutor;
    private ConcurrentTaskExecutor executor;

    /**
     * Creates a new work queue with the given max workers.<p>
     * If the max workers is greater than one, the provider work executor must be thread safe.
     *
     * @param name    Symbolic name for the work queue
     * @param workers Maximum workers allowed to work on this queue
     * @param target  The target object to execute the method invocation
     */
    public WorkQueueImpl(String name, int workers, Object target, @Nonnull ConflictsGuard<String> conflictsGuard) {
        this(name, workers, target, null, conflictsGuard);
    }

    /**
     * Creates a new work queue with the given max workers.<p>
     * If the max workers is greater than one, the provider work executor must be thread safe.
     *
     * @param name           Symbolic name for the work queue
     * @param workers        Maximum workers allowed to work on this queue
     * @param target         The target object to execute the method invocation
     * @param customExecutor custom executor
     * @param conflictsGuard custom conflicts guard
     */
    public WorkQueueImpl(String name, int workers, Object target,
            @Nullable AsyncWorkQueueCustomExecutor customExecutor,
            @Nonnull ConflictsGuard<String> conflictsGuard) {
        this.name = name;
        this.workers = workers;
        this.target = target;
        this.semaphore = new Semaphore(workers);
        this.queue = new NonBlockingOnWriteQueue<>(conflictsGuard);
        this.customExecutor = customExecutor;
    }

    @Override
    public boolean offerWork(T workItem, Method method) {
        log.trace("adding {}: to '{}'", workItem, name);
        boolean added = queue.addToPending(workItem, method);
        if (!added) {
            log.trace("{}: already contains '{}'", name, workItem);
        } else {
            log.trace("{}: successfully added to '{}'", workItem, name);
        }
        return added;
    }

    @Override
    public int availablePermits() {
        return semaphore.availablePermits();
    }

    @Override
    public boolean tryAcquire() {
        log.trace("Acquiring semaphore: {}", getName());
        return semaphore.tryAcquire();
    }

    @Override
    public void release() {
        log.trace("Releasing semaphore: {}", getName());
        semaphore.release();
    }

    @Override
    public Map<String, Object> doJobs() {
        boolean shouldPromote = true;
        try {
            debugStep("start processing");
            WorkQueuePromotedItem<T> promotedWorkItem;
            Map<String, Object> result = new HashMap<>();
            Instant start = Instant.now();

            while ((shouldPromote = shouldPromote(start)) && (promotedWorkItem = queue.promote()) != null) {
                T workItem = promotedWorkItem.workItem;
                traceStep("started working", workItem);
                try {
                    Object object = invoke(promotedWorkItem, workItem);
                    if (object != null) {
                        result.put(workItem.getUniqueKey(), object);
                    }
                    traceStep("finished working", workItem);
                } catch (Exception e) {
                    log.error("{}: failed to process {}", name, workItem, e);
                } finally {
                    boolean removed = queue.remove(promotedWorkItem);
                    if (removed) {
                        traceStep("successfully removed", workItem);
                    } else {
                        log.error("unexpected state: failed to remove {} from {}. Queue size pending={} running={}",
                                workItem, name, queue.getQueueSize(), queue.getRunningSize());
                    }
                }
            }
            return result;
        } catch (Exception e) {
            log.warn("Error occurred while promoting queue items. {}", e.getMessage());
            log.debug("Error occurred while promoting queue items.", e);
            throw new RuntimeException(e);
        } finally {
            release();
            //if promotion stopped and has more items in the queue than re-submit a task
            if (!shouldPromote && getQueueSize() > 0) {
                log.debug("Work queue {} is running more than {} sec, stop and re-submit", name,
                        ConstantValues.workQueueDoJobIntervalSecs);

                execute(executor());
            }
        }
    }

    @Override
    public void execute(ConcurrentTaskExecutor executor) {
        // Limit number of workers by semaphore
        // Submit only can acquire, to avoid executor overflow
        if (tryAcquire()) {
            try {
                executor.submit(this::doJobs);
            } catch (RejectedExecutionException e) {
                release();
                log.error("Failed to submit work queue {} task", getName(), e);
            }
        } else {
            log.debug("{}: max workers already processing ({})", getName(), getMaxNumberOfWorkers());
        }
    }

    /**
     * We don't want to let work queue run for long time if there are pending tasks to execution.
     * Should promote if duration is less than ConstantValues.workQueueMaxExecutionSec or NO pending tasks
     */
    @VisibleForTesting
    boolean shouldPromote(Instant start) {
        if (Duration.between(start, Instant.now()).toSeconds() > ConstantValues.workQueueMaxExecutionSec.getLong() &&
                ((ArtifactoryConcurrentExecutor) executor().getConcurrentExecutor()).getPendingTasks() > 0) {
            return false;
        }
        return true;
    }

    private ConcurrentTaskExecutor executor() {
        if (customExecutor != null && customExecutor.getExecutor() != null) {
            return customExecutor.getExecutor();
        }

        if (executor == null) {
            synchronized (this) {
                if (executor == null) {
                    executor = ContextHelper.get().beanForType(CachedThreadPoolTaskExecutor.class);
                }
            }
        }
        return executor;
    }


    private Object invoke(WorkQueuePromotedItem<T> promotedWorkItem, T workItem) {
        try {
            return promotedWorkItem.method.invoke(target, workItem);
        } catch (Exception e) {
            String origMsg = e.getMessage();
            if (e instanceof InvocationTargetException && e.getCause() != null) {
                origMsg = e.getCause().getMessage();
            }
            String msg = "Failed to call work queue '" + name + "' callback due to :" + origMsg;
            log.error(msg, e);
            throw new RuntimeException(msg, e);
        }
    }

    private void debugStep(String stepName) {
        // queue size is expensive, and returns current running tasks from other nodes as well in HA.
        if (log.isDebugEnabled()) {
            log.debug("{}: {}. Queue size pending={} running={}", name, stepName, queue.getQueueSize(),
                    queue.getRunningSize());
        }
    }

    private void traceStep(String stepName, T workItem) {
        // queue size is expensive, and returns current running tasks from other nodes as well in HA.
        if (log.isTraceEnabled()) {
            log.trace("{}: {} on {}. Queue size pending={} running={}", name, stepName, workItem, queue.getQueueSize(),
                    queue.getRunningSize());
        }
    }

    @Override
    public void stopQueue() {
        queue.stop();
    }

    @Override
    public int getQueueSize() {
        return queue.getQueueSize();
    }

    @Override
    public int getNumberOfWorkers() {
        return getMaxNumberOfWorkers() - semaphore.availablePermits();
    }

    @Override
    public int getMaxNumberOfWorkers() {
        return workers;
    }

    @Override
    public String getName() {
        return name;
    }

    @Override
    public void waitForItemDone(T workItem) {
        if (queue.contains(workItem)) {
            long timeout = ConstantValues.workQueueSyncExecutionTimeoutMillis.getLong();
            // Pending or running => Wait on work item
            synchronized (workItem) {
                if (queue.contains(workItem)) {
                    try {
                        workItem.wait(timeout);
                    } catch (InterruptedException e) {
                        throw new RuntimeException(
                                "Work Item " + workItem + " did not finished completion in " + timeout + "ms", e);
                    }
                }
            }
        }
    }

    @Override
    public AsyncWorkQueueCustomExecutor customExecutor() {
        return customExecutor;
    }

    @Override
    public void updateWorkers(int newWorkers) throws InterruptedException {
        if (newWorkers != workers) {
            log.info("Updating {} workers: {} -> {}", name, workers, newWorkers);
            if (newWorkers > workers) {
                semaphore.release(newWorkers - workers);
            } else {// if (newWorkers < workers)
                // reduce permits one by one, so we will not have to wait for all if the reduction count is used.
                for (int index = 1; index <= workers - newWorkers; index++) {
                    semaphore.acquire();
                }
            }
            workers = newWorkers;
        }
    }
}
