/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.artifactory.traffic.v2;

import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Getter;
import lombok.NoArgsConstructor;
import lombok.Setter;
import org.apache.commons.lang.StringUtils;
import org.artifactory.traffic.XrayTrafficLogger;
import org.artifactory.traffic.entry.TransferEntry;
import org.jfrog.storage.binstore.providers.base.RedirectType;
import org.joda.time.format.DateTimeFormat;
import org.joda.time.format.DateTimeFormatter;
import org.slf4j.MDC;

import static org.apache.commons.lang.StringUtils.startsWith;

/**
 * Entry for data transfer traffic entries.
 * <p/>
 * $date|$traceId|$action|$actor|$redirectTarget|$remoteAddress|$repoPath|$size|$requestedSize|$transferredSize
 * Example:
 * <pre>
 * 20200608005250|2de5e67cc72235bb|DOWNLOAD|USER|DIRECT|123.123.123.123|
 * nuget-local:Application/Seismic.LiveDoc.Engine.6.12.0.nupkg|9274323|8000000|7368479
 *
 * 20200608005250|2de5e67cc72235bb|REDIRECT|XRAY|S3|123.123.123.123|
 * nuget-local:Application/Seismic.LiveDoc.Engine.6.12.0.nupkg|9274323|9274323|9274323
 * </pre>
 *
 * @author Rotem Kfir
 */
@Getter
@Builder
@NoArgsConstructor
@AllArgsConstructor
public class TrafficEntryV2 {
    private static final DateTimeFormatter ENTRY_DATE_FORMATTER = DateTimeFormat.forPattern("yyyyMMddHHmmss");

    private String traceId;
    private TrafficAction action;
    private TrafficActor actor;
    private TrafficTarget redirectTarget;
    private String userAddress;
    private String repoPath;
    /**
     * The full size (in bytes) of the blob
     */
    private long size;
    /**
     * The number of bytes requested, e.g. in a range request, or the full size
     */
    private long requestedSize;
    /**
     * The actual number of bytes transferred
     */
    private long transferredSize;

    @Setter
    private String projectKey;

    @Setter
    private String repoKey;

    private String getFormattedDate() {
        return ENTRY_DATE_FORMATTER.print(System.currentTimeMillis());
    }

    protected String getTraceId() {
        if (traceId != null) {
            return traceId;
        }
        return MDC.get("uber-trace-id");
    }

    @Override
    public String toString() {
        return String.join("|",
                getFormattedDate(), getTraceId(), getAction().name(), getActor().name(),
                getRedirectTarget() == null ? "" : getRedirectTarget().name(),
                getUserAddress(), getRepoPath(), String.valueOf(getSize()), String.valueOf(getRequestedSize()),
                String.valueOf(getTransferredSize()));
    }

    public String toStringIncludingProjectAndRepo() {
        return String.join("|", toString(), getRepoKey(), getProjectKey());
    }

    public enum TrafficAction {
        DOWNLOAD, UPLOAD, REDIRECT
    }

    /**
     * The type of the actor who initiated the request
     */
    public enum TrafficActor {
        USER, XRAY, REPLICATOR, PIPELINES, FEDERATION, ARTIFACTORY, OBSERVABILITY, JFCONNECT, OBSERVABILITY_USAGE;

        public static TrafficActor valueOfIdentifier(String identifier, boolean xrayAuthenticationUser) {
            if (StringUtils.isNotBlank(identifier)) {
                if (xrayAuthenticationUser) {
                    return XRAY;
                }
                if (startsWith(identifier.toLowerCase(), "jfrog-replicator/")) {
                    return REPLICATOR;
                }
            }
            return USER;
        }

    }

    /**
     * The target type of the transfer
     */
    public enum TrafficTarget {
        DIRECT, // in case the action is not REDIRECT
        S3, // redirect to Amazon S3
        CF, // redirect to Amazon CloudFront CDN
        GCS, // redirect to Google Cloud Service
        ABS, // redirect to Azure Blob Storage
        ;

        public static TrafficTarget valueOf(RedirectType redirectType) {
            if (redirectType == null) {
                return DIRECT;
            }
            switch (redirectType) {
                case S3:
                    return S3;
                case CLOUD_FRONT:
                    return CF;
                case GCS:
                    return GCS;
                case ABS:
                    return ABS;
                default:
                    return DIRECT;
            }
        }
    }

    public static TrafficEntryV2 map(TransferEntry entry) {
        TrafficEntryV2.TrafficAction trafficAction;

        switch (entry.getAction()) {
            case UPLOAD:
                trafficAction = TrafficEntryV2.TrafficAction.UPLOAD;
                break;
            case REDIRECT:
                trafficAction = TrafficEntryV2.TrafficAction.REDIRECT;
                break;
            default:
                trafficAction = TrafficEntryV2.TrafficAction.DOWNLOAD;
        }

        return TrafficEntryV2.builder()
                .action(trafficAction)
                .actor(
                        TrafficActor.valueOfIdentifier(
                                entry.getUserIdentifier(),
                                XrayTrafficLogger.isXrayTraffic(entry)
                        )
                )
                .redirectTarget(null)
                .repoPath(entry.getRepoPath())
                .userAddress(entry.getUserAddress())
                .traceId(entry.getTraceId())
                .requestedSize(entry.getContentLength())
                .size(entry.getContentLength())
                .transferredSize(entry.getContentLength())
                .projectKey("")
                .repoKey("")
                .build();
    }
}
