/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.aql.sql.builder.query.aql;

import org.artifactory.aql.AqlException;
import org.artifactory.aql.model.AqlComparatorEnum;
import org.artifactory.aql.model.AqlDomainEnum;
import org.artifactory.aql.model.AqlValue;
import org.artifactory.aql.model.AqlVariable;
import org.artifactory.storage.db.aql.sql.builder.query.sql.SqlTable;
import org.jetbrains.annotations.NotNull;

import java.util.List;

import static org.artifactory.storage.db.aql.sql.builder.query.sql.type.AqlTableGraph.getFromFieldToProperties;
import static org.artifactory.storage.db.aql.sql.builder.query.sql.type.AqlTableGraph.getLinkTable;

/**
 * @author Gidi Shabat
 */
public class ComplexPropertyCriterionSubSelect extends Criterion {

    public ComplexPropertyCriterionSubSelect(List<AqlDomainEnum> subDomains, AqlVariable variable1, SqlTable table1,
            String comparatorName, AqlVariable variable2, SqlTable table2, boolean mspOperator) {
        super(subDomains, variable1, table1, comparatorName, variable2, table2, mspOperator);
    }

    public ComplexPropertyCriterionSubSelect(Criterion criterion) {
        super(criterion.getSubDomains(), new AqlValue(((AqlValue) criterion.getVariable1()).getValueType(),
                        ((AqlValue) criterion.getVariable1()).getValue()), criterion.getTable1(),
                criterion.getComparatorName(), new AqlValue(((AqlValue) criterion.getVariable2()).getValueType(),
                        ((AqlValue) criterion.getVariable2()).getValue()),
                criterion.getTable2(), criterion.isMspOperator());
    }

    /**
     * Converts propertyCriteria to Sql criteria
     */
    @Override
    public String toSql(List<Object> params) throws AqlException {
        // Get both variable which are Values (this is property criteria)
        AqlVariable value1 = getVariable1();
        AqlVariable value2 = getVariable2();
        // Get both tables which are node_props tables (this is property criteria)
        SqlTable table1 = getTable1();
        // SqlTable table2 = getTable2();
        // update the Sql input param list
        // Get the ComparatorEnum
        AqlComparatorEnum comparatorEnum = AqlComparatorEnum.value(getComparatorName());
        return createSqlComplexPropertyCriteria(comparatorEnum, value1, table1, value2, params);
    }

    @Override
    public String toNative() {
        StringBuilder builder = new StringBuilder();
        builder.append("{\"").append(getPath(this.getSubDomains(), this.getSubDomains().size() - 1))
                .append("@").append(this.getVariable1().toNative()).append("\":{");
        builder.append("\"").append(this.getComparatorName()).append("\":");
        String value = this.getVariable2().toNative();
        if (value.equals("null")) {
            builder.append(value);
        } else {
            builder.append("\"").append(value).append("\"");
        }
        builder.append("}}");
        return builder.toString();
    }

    private String createSqlComplexPropertyCriteria(AqlComparatorEnum comparatorEnum, AqlVariable variable1,
            SqlTable table1,
            AqlVariable variable2, List<Object> params) {
        String tableName = table1.getTableName();
        String fieldName = getFromFieldToProperties(table1.getTable()).name();

        String subSelectAlias = getSubSelectAlias();
        String shortAlias = getTable1().getTable().alias + "sub.";
        StringBuilder where = new StringBuilder("( ");
        where.append(subSelectAlias).append(fieldName).append(" not in ").append("( ").append("select ")
                .append(fieldToSql(shortAlias, fieldName)).append(" from ")
                .append(tableName).append(" ").append(getTable1().getTable().alias + "sub").append(" where ")
                .append(fieldToSql(shortAlias, fieldName)).append(" = ").append(subSelectAlias)
                .append(fieldName).append(" and ").append(fieldToSql(shortAlias, "prop_key"))
                .append(" = ").append(PARAMETER_PLACEHOLDER).append(" and ")
                .append(fieldToSql(shortAlias, "prop_value"));
        switch (comparatorEnum) {
            case notMatches: {
                Object param1 = resolveParam(variable1);
                String param2 = resolveParamForSqlLikeQuery(variable2);
                params.add(param1);
                params.add(param2);
                where.append(" Like ")
                        .append(PARAMETER_PLACEHOLDER).append(generateEscapingCriteria(param2)).append(" )) ");
                return where.toString();
            }
            case notEquals: {
                Object param1 = resolveParam(variable1);
                Object param2 = resolveParam(variable2);
                params.add(param1);
                params.add(param2);
                where.append(" = ")
                        .append(PARAMETER_PLACEHOLDER).append(" )) ");
                return where.toString();
            }
            default:
                throw new IllegalStateException("Should not reach to the point of code");
        }
    }


    @NotNull
    private String getSubSelectAlias() {
        // if only one table in subdomain it means we need to have the alias of the prop table, else we need the one
        // table prior props
        return (getSubDomains().size() == 1) ?
                getLinkTable(getTable1().getTable(), getSubDomains().get(getSubDomains().size() - 1)).getAlias() :
                getLinkTable(getTable1().getTable(), getSubDomains().get(getSubDomains().size() - 2)).getAlias();
    }
}
