/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2019 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.binstore.service;

import com.google.common.base.Suppliers;
import com.google.common.collect.Lists;
import com.google.common.collect.MapMaker;
import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.io.IOUtils;
import org.apache.commons.io.output.NullOutputStream;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.ClusterOperationsService;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.addon.ha.interceptor.ClusterTopologyListener;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.storage.BinariesInfo;
import org.artifactory.binary.headers.BinaryHeaders;
import org.artifactory.checksum.ChecksumInfo;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.checksum.ChecksumsInfo;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.ha.HaNodeProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.concurrent.NonStandardLockCategory;
import org.artifactory.config.CentralConfigKey;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.environment.BinaryStoreProperties;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.Reloadable;
import org.artifactory.storage.BinaryInsertRetryException;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.StorageException;
import org.artifactory.storage.binstore.service.BinariesGarbageCollectorService;
import org.artifactory.storage.binstore.service.BinaryService;
import org.artifactory.storage.binstore.service.GarbageCollectionDatastoreService;
import org.artifactory.storage.binstore.service.GarbageCollectorInfo;
import org.artifactory.storage.binstore.service.GarbageCollectorListener;
import org.artifactory.storage.binstore.service.GarbageCollectorStrategy;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.artifactory.storage.db.InternalDbService;
import org.artifactory.storage.db.binstore.dao.BinariesDao;
import org.artifactory.storage.db.binstore.entity.BinaryEntity;
import org.artifactory.storage.db.binstore.exceptions.PruneException;
import org.artifactory.storage.db.binstore.util.BinaryServiceUtils;
import org.artifactory.storage.db.binstore.util.FilestorePruner;
import org.artifactory.storage.db.binstore.visitors.BinaryTreeElementScanner;
import org.artifactory.storage.db.binstore.visitors.EssentialBinaryTreeElementHandler;
import org.artifactory.storage.db.remotebintask.RemoteBinaryTaskDao;
import org.artifactory.storage.db.remotebintask.entity.RemoteBinaryTask;
import org.artifactory.storage.db.servers.model.ArtifactoryServer;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.artifactory.storage.fs.service.ArchiveEntriesService;
import org.artifactory.storage.model.FileBinaryProviderInfo;
import org.artifactory.util.AolUtils;
import org.artifactory.util.CollectionUtils;
import org.artifactory.util.StreamUtils;
import org.artifactory.version.CompoundVersionDetails;
import org.jetbrains.annotations.NotNull;
import org.jfrog.access.client.AccessClient;
import org.jfrog.common.concurrency.ConflictGuard;
import org.jfrog.common.concurrency.ConflictsGuard;
import org.jfrog.common.concurrency.LockingMapFactoryProvider;
import org.jfrog.security.util.Pair;
import org.jfrog.storage.DbType;
import org.jfrog.storage.authentication.AuthenticationTokenProvider;
import org.jfrog.storage.binstore.common.BinaryElementImpl;
import org.jfrog.storage.binstore.common.BinaryElementRequestImpl;
import org.jfrog.storage.binstore.common.BinaryStoreContextImpl;
import org.jfrog.storage.binstore.common.ChecksumInputStream;
import org.jfrog.storage.binstore.common.ReaderTrackingInputStream;
import org.jfrog.storage.binstore.exceptions.BinaryNotFoundException;
import org.jfrog.storage.binstore.exceptions.BinaryRejectedException;
import org.jfrog.storage.binstore.exceptions.BinaryStorageException;
import org.jfrog.storage.binstore.exceptions.NoCloudProviderConfiguredException;
import org.jfrog.storage.binstore.exceptions.SignedUrlException;
import org.jfrog.storage.binstore.ifc.BinaryProviderConfig;
import org.jfrog.storage.binstore.ifc.BinaryProviderInfo;
import org.jfrog.storage.binstore.ifc.BinaryStoreFederatedMember;
import org.jfrog.storage.binstore.ifc.BinaryStream;
import org.jfrog.storage.binstore.ifc.ClusterNode;
import org.jfrog.storage.binstore.ifc.ProviderConnectMode;
import org.jfrog.storage.binstore.ifc.UsageTracking;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.model.BinaryTreeElement;
import org.jfrog.storage.binstore.ifc.model.StorageInfo;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.manager.BinaryProviderManagerImpl;
import org.jfrog.storage.binstore.providers.base.BinaryProviderBase;
import org.jfrog.storage.binstore.providers.base.CloudBinaryProvider;
import org.jfrog.storage.binstore.providers.base.ManualRedirectType;
import org.jfrog.storage.binstore.providers.base.PersistedRemoteTask;
import org.jfrog.storage.binstore.providers.base.PersistedTaskManager;
import org.jfrog.storage.binstore.providers.base.RedirectType;
import org.jfrog.storage.binstore.utils.Checksum;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Supplier;
import java.util.stream.Collectors;

import static org.apache.commons.lang.StringUtils.isBlank;
import static org.apache.commons.lang.StringUtils.isNotBlank;
import static org.artifactory.checksum.ChecksumInfo.TRUSTED_FILE_MARKER;
import static org.artifactory.state.ArtifactoryServerState.RUNNING;
import static org.artifactory.storage.db.binstore.util.BinaryServiceUtils.collectBinaryProviderInfo;
import static org.artifactory.storage.db.binstore.util.BinaryServiceUtils.collectFileBinaryProvidersDirsInternal;
import static org.artifactory.storage.db.binstore.util.BinaryServiceUtils.isDuplicatedEntryException;
import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.CHECKSUM_MD5;
import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.CHECKSUM_SHA1;
import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.CHECKSUM_SHA256;
import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.LOCAL_ONLY;

/**
 * The main binary store of Artifactory that delegates to the BinaryProvider chain.
 *
 * @author Yossi Shaul
 */
@Slf4j
@Service
@Reloadable(beanClass = BinaryService.class,
        initAfter = {AccessService.class, ClusterOperationsService.class},
        listenOn = CentralConfigKey.none)
public class BinaryServiceImpl implements InternalBinaryService, ClusterTopologyListener {

    private final AtomicInteger gcIteration = new AtomicInteger(0);
    private final List<String> s3Providers = List.of("state-aware-s3", "s3-storage-v3");
    private final Lock lock = new ReentrantLock();
    private final List<GarbageCollectorListener> garbageCollectorListeners = new CopyOnWriteArrayList<>();
    private BinariesDao binariesDao;
    private RemoteBinaryTaskDao remoteBinaryTaskDao;
    private AccessService accessService;
    private ArchiveEntriesService archiveEntriesService;
    private GarbageCollectionDatastoreService garbageCollectionDatastoreService;
    private InternalDbService dbService;
    private DeleteBlobInfosWrapper deleteBlobInfosWrapper;
    private ArtifactoryTracer artifactoryTracer;

    /**
     * Map of delete protected sha1 checksums to the number of protections (active readers + writer count for each
     * binary)
     */
    private ConcurrentMap<String, Pair<AtomicInteger, Long>> deleteProtectedBinaries;
    private BinaryProviderManager binaryProviderManager;
    private BinaryProviderConfig defaultValues;
    private AuthenticationTokenProvider authenticationTokenProvider;

    @Autowired
    public void setBinariesDao(BinariesDao binariesDao) {
        this.binariesDao = binariesDao;
    }

    @Autowired
    public void setRemoteBinaryTaskDao(RemoteBinaryTaskDao remoteBinaryTaskDao) {
        this.remoteBinaryTaskDao = remoteBinaryTaskDao;
    }

    @Autowired
    public void setArchiveEntriesService(ArchiveEntriesService archiveEntriesService) {
        this.archiveEntriesService = archiveEntriesService;
    }

    @Autowired
    public void setGarbageCollectionDatastoreService(
            GarbageCollectionDatastoreService garbageCollectionDatastoreService) {
        this.garbageCollectionDatastoreService = garbageCollectionDatastoreService;
    }

    @Autowired
    public void setAccessService(AccessService accessService) {
        this.accessService = accessService;
    }

    @Autowired
    public void setAuthenticationTokenProvider(AuthenticationTokenProvider authenticationTokenProvider) {
        this.authenticationTokenProvider = authenticationTokenProvider;
    }

    @PostConstruct
    public void initialize() {
        log.debug("Initializing the ConfigurableBinaryProviderManager");
        deleteProtectedBinaries = new MapMaker().makeMap();
        // Generate Default values
        ArtifactoryHome artifactoryHome = ArtifactoryHome.get();
        BinaryStoreProperties storeProperties = new BinaryStoreProperties(artifactoryHome.getDataDir().getPath(),
                artifactoryHome.getSecurityDir().getPath());
        HaNodeProperties haNodeProperties = artifactoryHome.getHaNodeProperties();
        if (artifactoryHome.isHaConfigured() && haNodeProperties != null) {
            storeProperties.setClusterDataDir(haNodeProperties.getClusterDataDir());
        }
        defaultValues = storeProperties.toDefaultValues();

        PersistedTaskManager<PersistedRemoteTask> persistedTaskManager =
                new PersistedTaskManagerImpl(remoteBinaryTaskDao, this, dbService);
        // TODO: [by fsi] this ha enabled needs to be done at the binary store project level
        if (artifactoryHome.isHaConfigured() && haNodeProperties != null) {
            defaultValues.setBinaryStoreContext(new BinaryStoreContextImpl(new ArtifactoryLockingMapFactoryProvider(),
                    haNodeProperties.getProperties(), this::exists, this::getAccessClient,
                    artifactoryTracer::getFullContextTracingId, persistedTaskManager, this::getTokenProvider,
                    getNumOfActiveServersSupplier(),
                    ConstantValues.binaryStoreFullBinaryProviderChainPing.getBoolean()));
            defaultValues.addParam("serviceId", haNodeProperties.getServerId() + "-binary-store");
        } else {
            // The only difference here is that we don't init the HaNodeProperties
            defaultValues.setBinaryStoreContext(
                    new BinaryStoreContextImpl(new ArtifactoryLockingMapFactoryProvider(), null, this::exists,
                            this::getAccessClient, artifactoryTracer::getFullContextTracingId, persistedTaskManager,
                            this::getTokenProvider, getNumOfActiveServersSupplier(),
                            ConstantValues.binaryStoreFullBinaryProviderChainPing.getBoolean()));
        }

        // Set the binarystore.xml file location
        File haAwareEtcDir = artifactoryHome.getEtcDir();
        File userConfigFile = new File(haAwareEtcDir, "binarystore.xml");
        defaultValues.setBinaryStoreXmlPath(userConfigFile.getPath());
        // Finally create an instance of the binary persistedTaskManager manager
        binaryProviderManager = new BinaryProviderManagerImpl(defaultValues, artifactoryHome.
                getJFrogMasterKeyEncrypter());
    }

    @Autowired
    public void setDbService(InternalDbService dbService) {
        this.dbService = dbService;
    }

    @Autowired
    public void setDeleteBlobInfosWrapper(DeleteBlobInfosWrapper deleteBlobInfosWrapper) {
        this.deleteBlobInfosWrapper = deleteBlobInfosWrapper;
    }

    @Override
    public BinaryProviderManager getBinaryProviderManager() {
        return binaryProviderManager;
    }

    @Override
    @PreDestroy
    public void destroy() {
        notifyGCListenersOnDestroy();
        binaryProviderManager.contextDestroyed();
    }

    @Override
    public void addGCListener(GarbageCollectorListener garbageCollectorListener) {
        garbageCollectorListeners.add(garbageCollectorListener);
    }

    @Override
    public void addExternalFileStore(File externalFileDir, ProviderConnectMode connectMode) {
        // The external binary provider works only if the file binary provider is not null
        if (getBinariesDir() == null) {
            return;
        }
        // Prepare parameters for the new External binary provider
        String mode = connectMode.propName;
        String externalDir = externalFileDir.getAbsolutePath();
        String fileStoreDir = defaultValues.getParam("fileStoreDir");
        File fileStoreFullPath = new File(new File(defaultValues.getParam("baseDataDir")), fileStoreDir);
        // create and initialize the external binary providers.
        binaryProviderManager.initializeExternalBinaryProvider(mode, externalDir, fileStoreFullPath.getAbsolutePath(),
                defaultValues);
    }

    @Override
    public void disconnectExternalFilestore(File externalDir, ProviderConnectMode disconnectMode,
            BasicStatusHolder statusHolder) {
        ExternalBinaryProviderHelper.disconnectFromFileStore(this, externalDir, disconnectMode, statusHolder,
                binaryProviderManager, binariesDao, defaultValues);
    }

    @Override
    public File getBinariesDir() {
        // Get binary providers info tree from the manager
        BinaryProvidersInfo<BinaryProviderInfo> binaryProvidersInfo = binaryProviderManager.getBinaryProvidersInfo();
        BinaryTreeElement<BinaryProviderInfo> treeElement = binaryProvidersInfo.rootTreeElement;
        // Collect all the file binary providers in list
        List<FileBinaryProviderInfo> providersInfos = Lists.newArrayList();
        collectFileBinaryProvidersDirsInternal(providersInfos, treeElement);
        // Get the First binary provider
        FileBinaryProviderInfo fileBinaryProviderInfo = providersInfos.size() > 0 ? providersInfos.get(0) : null;
        if (fileBinaryProviderInfo != null) {
            // We need the wrapper to avoid binary dir recalculation even if there is no file binary provider
            return fileBinaryProviderInfo.getFileStoreDir();
        }
        return null;
    }

    @Override
    public boolean isCloudProviderSupportingRedirectionConfigured() {
        List<CloudBinaryProvider> cloudBinaryProvider = getCloudBinaryProviders();
        if (cloudBinaryProvider.isEmpty()) {
            return false;
        }
        return cloudBinaryProvider.stream().allMatch(CloudBinaryProvider::supportsRedirection);
    }

    private List<BinaryTreeElement<BinaryProviderInfo>> getBinaryTreeElementsList() {
        BinaryProvidersInfo<BinaryProviderInfo> binaryProvidersInfo = binaryProviderManager.getBinaryProvidersInfo();
        List<BinaryTreeElement<BinaryProviderInfo>> providersInfos = Lists.newArrayList();
        collectBinaryProviderInfo(providersInfos, binaryProvidersInfo.rootTreeElement);
        return providersInfos;
    }

    @Override
    public StorageInfo getStorageInfoSummary() {
        List<BinaryTreeElement<BinaryProviderInfo>> providersInfos = getBinaryTreeElementsList();
        // Quota is for the final NFS if exists, but if final is infinite quota should be for cache fs binary provider
        // Take the smallest space available from all the providers that are not cache fs
        StorageInfo cacheFsStorageInfo = null;
        StorageInfo smallestStorageInfo = null;
        for (BinaryTreeElement<BinaryProviderInfo> providerInfo : providersInfos) {
            if (providerInfo != null) {
                BinaryProviderInfo data = providerInfo.getData();
                StorageInfo storageInfo = data.getStorageInfo();
                String type = data.getProperties().get("type");
                if ("cache-fs".equals(type)) {
                    cacheFsStorageInfo = storageInfo;
                    // An EFS mount generates negative free space!
                    if (cacheFsStorageInfo.getFreeSpace() < 0L || cacheFsStorageInfo.getFreeSpaceInPercent() < 0L ||
                            cacheFsStorageInfo.getFreeSpaceInPercent() > 100L) {
                        // It's infinite
                        cacheFsStorageInfo = BinaryProviderBase.INFINITE_STORAGE_INFO;
                    }
                } else if (!BinaryProviderBase.EMPTY_STORAGE_INFO.equals(storageInfo)) {
                    if (smallestStorageInfo == null) {
                        smallestStorageInfo = storageInfo;
                    } else {
                        // An EFS mount generates negative free space!
                        if (storageInfo.getFreeSpaceInPercent() > 0L &&
                                smallestStorageInfo.getFreeSpaceInPercent() > storageInfo.getFreeSpaceInPercent()) {
                            // TODO: [by fsi] May be have the caller aware of many layers?
                            smallestStorageInfo = storageInfo;
                        }
                    }
                }
            }
        }
        if (smallestStorageInfo != null) {
            if (BinaryProviderBase.INFINITE_STORAGE_INFO.equals(smallestStorageInfo) ||
                    smallestStorageInfo.getFreeSpace() == Long.MAX_VALUE || smallestStorageInfo.getFreeSpace() == -1L) {
                // If there is a chae fs above the infinite storage use this to check quota
                if (cacheFsStorageInfo != null) {
                    return cacheFsStorageInfo;
                }
            }
            return smallestStorageInfo;
        }
        // No smallest found, if cache use it.
        return Objects.requireNonNullElse(cacheFsStorageInfo, BinaryProviderBase.UNKNOWN_STORAGE_INFO);
        // No storage info relevant found, return unknown
    }

    @NonNull
    public List<BinaryElement> getBinaryRecords(@NonNull List<String> sha1s) {
        try {
            List<BinaryEntity> binaryEntities = binariesDao.loadBySha1s(sha1s);
            Map<String, BinaryEntity> sha1Map = new HashMap<>();
            binaryEntities.forEach(entity -> {
                String sha1 = entity.getSha1();
                sha1Map.put(sha1, entity);
            });

            return sha1s.stream().map(sha1 -> {
                BinaryEntity binaryEntity = sha1Map.get(sha1);
                if (binaryEntity != null) {
                    return convertToBinaryElement(binaryEntity);
                } else {
                    return null;
                }
            }).collect(Collectors.toList());

        } catch (SQLException e) {
            throw new StorageException(e);
        }
    }

    @Override
    @Nullable
    public BinaryElement addBinaryRecord(Map<String, String> headers, String sha1, String sha2, String md5,
            long length) {
        try {
            //[sha2]: switch to sha2 when storage supports it
            BinaryEntity result = binariesDao.load(ChecksumType.sha1, sha1);
            if (result == null) {
                // It does not exists in the DB
                // Let's check if in bin provider
                BinaryElementRequestImpl request = new BinaryElementRequestImpl(sha1, sha2, md5, length);
                request.getHeaders().putAll(headers);
                if (binaryProviderManager.getFirstBinaryProvider().exists(request).exists()) {
                    //info missing? pull binary from the store and recalculate checksums
                    sha2 = fillInMissingChecksumInfo(sha1, sha2, headers);
                    //sanity!
                    if (isBlank(sha1) || isBlank(sha2) || isBlank(md5)) {
                        log.debug("Missing checksum information for binary with SHA1 '{}', SHA256 '{}', MD5 '{}'. " +
                                "It will not be added", sha1, sha2, md5);
                        return null;
                    }
                    // Good let's use it
                    return getTransactionalMe().insertRecordInDb(sha1, sha2, md5, length);
                }
                return null;
            }
            return convertToBinaryElement(result);
        } catch (SQLException e) {
            throw new StorageException("Could not reserve entry '" + sha1 + "'", e);
        }
    }

    /**
     * Used by {@link #addBinaryRecord} to complete missing sha2 data for any action the tries writing binary records
     * using already-existing binaries (i.e. checksum deploy, import
     */
    private String fillInMissingChecksumInfo(String sha1, String sha2, Map<String, String> headers) {
        //when we switch checksum type in the filestore a check for all checksum types must
        //be added here... for now the only possible scenario is export from version < 5.5.0
        //and import into version >= 5.5.0 which means there's no sha2 metadata for the file
        if (isNotBlank(sha2)) {
            return sha2;
        }
        //For import scenarios it would have been better to pull checksums from the source file, but for
        //the sake of over-complexity i'm doing a pull from the filestore.
        BinaryElement element = createBinaryElement(sha1, null, null, -1, headers);
        element.addHeader(BinaryElementHeaders.SKIP_CACHE, "true");
        Checksum sha2Checksum = new Checksum(ChecksumType.sha256.alg(), ChecksumType.sha256.length());
        try (InputStream in = getBinary(element); ChecksumInputStream calcStream =
                new ChecksumInputStream(in, sha2Checksum)) {
            //waste the stream to have it calculate the checksum
            IOUtils.copy(calcStream, new NullOutputStream());
            calcStream.close();
            sha2 = sha2Checksum.getChecksum();
        } catch (Exception e) {
            log.error("Can't calculate missing checksum information for binary with SHA1 '{}' : {}", sha1,
                    e.getMessage());
            log.debug("", e);
        }
        return sha2;
    }

    @Override
    @Nonnull
    public BinaryElement addBinary(BinaryStream in) throws IOException, BinaryRejectedException {
        if (in.getStream() instanceof BinaryServiceInputStream) {
            throw new IllegalStateException(
                    "Cannot add binary from checksum deploy " + ((BinaryServiceInputStream) in).getBinaryInfo());
        }
        BinaryElement binaryInfo;
        BinaryElement bi = binaryProviderManager.getFirstBinaryProvider().addStream(in);
        log.trace("Inserted binary {} to file store", bi.getSha1());
        // From here we managed to create a binary record on the binary provider
        // So, failing on the insert in DB (because saving the file took to long)
        // can be re-tried based on the sha1
        try {
            binaryInfo = getTransactionalMe().insertRecordInDb(bi.getSha1(), bi.getSha2(), bi.getMd5(), bi.getLength());
        } catch (BinaryInsertRetryException e) {
            if (log.isDebugEnabled()) {
                log.info("Retrying add binary with SHA1: '" + bi.getSha1()
                        + "' and SHA256: '" + bi.getSha2() + "after receiving exception: ", e);
            } else {
                log.info("Retrying add binary with SHA1: '" + bi.getSha1()
                        + "' and SHA256: '" + bi.getSha2() + "after receiving exception: " + e.getMessage());
            }
            binaryInfo = addBinaryRecord(in.getHeaders(), bi.getSha1(), bi.getSha2(), bi.getMd5(), bi.getLength());
            if (binaryInfo == null) {
                throw new StorageException("Failed to add binary record with SHA1: '" + bi.getSha1()
                        + "' and SHA256: '" + bi.getSha2() + "' during retry", e);
            }
        }
        return binaryInfo;
    }

    @Override
    public BinaryProvidersInfo<Map<String, String>> getBinaryProvidersInfo() {
        // Get binary providers info tree from the binary store manager
        BinaryProvidersInfo<BinaryProviderInfo> binaryProvidersInfo = binaryProviderManager.getBinaryProvidersInfo();
        BinaryTreeElement<BinaryProviderInfo> teeElement = binaryProvidersInfo.rootTreeElement;
        // Create sub tree that contains only essential elements (for the UI)
        BinaryTreeElementScanner<BinaryProviderInfo, Map<String, String>> scanner = new BinaryTreeElementScanner<>();
        EssentialBinaryTreeElementHandler handler = new EssentialBinaryTreeElementHandler();
        BinaryTreeElement<Map<String, String>> scan = scanner.scan(teeElement, handler);
        return new BinaryProvidersInfo<>(binaryProvidersInfo.template, scan);
    }

    @Override
    public InputStream getBinary(String sha1, Map<String, String> headers) {
        final BinaryElementRequestImpl request = new BinaryElementRequestImpl(sha1);
        if (headers != null) {
            request.getHeaders().putAll(headers);
        }
        return new ReaderTrackingInputStream(binaryProviderManager.getFirstBinaryProvider().
                getStream(request), sha1, this);
    }

    @Override
    public InputStream getBinary(BinaryElement element) throws BinaryNotFoundException {
        //BE CAREFUL! users of this method currently supplies the sha1 value
        return new ReaderTrackingInputStream(binaryProviderManager.getFirstBinaryProvider().
                getStream(new BinaryElementRequestImpl(element)), element.getSha1(), this);
    }

    @Override
    public InputStream getBinaryForChecksumDeploy(BinaryElement element) throws BinaryNotFoundException {
        BinaryElementRequestImpl request = new BinaryElementRequestImpl(element);
        if (!binaryProviderManager.getFirstBinaryProvider().exists(request).exists()) {
            throw new BinaryNotFoundException("Binary provider has no content for " + element.getSha1() + " '");
        }
        return new BinaryServiceInputStreamWrapper(element, this);
    }

    @Override
    public RedirectType calcRedirectType(@NonNull RedirectType redirectType, boolean isAol,
            ManualRedirectType manualRedirectType) {
        List<CloudBinaryProvider> cloudProviders = getCloudBinaryProviders();
        if (cloudProviders.isEmpty()) {
            throw new NoCloudProviderConfiguredException("Couldn't calculate redirect type. " +
                    "There is no binary cloud provider configured. Check the logs for more details.");
        }
        validateProviders(cloudProviders);
        // One is enough
        return cloudProviders.get(0).calcRedirectType(redirectType, isAol, manualRedirectType);
    }

    @Override
    @NonNull
    public String getSignedUrl(@NonNull BinaryElement binaryElement) {
        String repoPathId = binaryElement.getHeaders().get(BinaryElementHeaders.REPO_PATH_ID.getHeaderName());
        List<CloudBinaryProvider> cloudProviders = getCloudBinaryProviders();
        if (cloudProviders.isEmpty()) {
            throw new SignedUrlException("Couldn't generate signed url for " + repoPathId + ". " +
                    "There is no binary cloud provider configured. Check the logs for more details.");
        }

        // If we are using s3-sharding we don't want to make an unnecessary api call to aws
        if (cloudProviders.size() == 1) {
            binaryElement.getHeaders().put("skipRemoteCheck", "true");
        }

        String signedUrl = null;
        for (CloudBinaryProvider provider : cloudProviders) {
            signedUrl = provider.getSignedUrl(binaryElement);
            if (signedUrl != null) {
                break;
            }
        }

        if (signedUrl == null) {
            throw new SignedUrlException("Couldn't generate signed url for " + repoPathId + ". " +
                    "Make sure " + cloudProviders + " binary providers is configured to support signed urls and check" +
                    "the logs for more details.");
        }
        log.trace("Generated signed url: {}", signedUrl);
        return signedUrl;
    }

    @Override
    public boolean canCloudProviderGenerateRedirection(String sha1) {
        List<CloudBinaryProvider> cloudProviders = getCloudBinaryProviders();
        if (cloudProviders.isEmpty()) {
            return false;
        }

        validateProviders(cloudProviders);
        return cloudProviders.stream()
                .anyMatch(curr -> curr.canGenerateRedirection(new BinaryElementRequestImpl(sha1)));
    }

    private void validateProviders(List<CloudBinaryProvider> cloudProviders) {
        String provider = cloudProviders.get(0).getProviderMetaData().getType().toLowerCase();
        cloudProviders.forEach(currProvider ->  {
            if (!currProvider.getProviderMetaData().getType().equalsIgnoreCase(provider)) {
                throw new BinaryStorageException("Not all cloud binary providers are of the same type");
            }
        });
    }

    private List<CloudBinaryProvider> getCloudBinaryProviders() {
        return binaryProviderManager.getCloudProviders();
    }

    @Override
    public boolean isCdnAllowed() {
        if (!isCloudProviderConfigured()) { return false; }
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        final List<CloudBinaryProvider> cloudProviders = binaryProviderManager.getCloudProviders();
        if (cloudProviders.isEmpty()) {
            return false;
        }
        boolean providerId;
        if (cloudProviders.size() == 1) {
            providerId = s3Providers.contains(cloudProviders.get(0).getProviderMetaData().getType().toLowerCase());
        } else {
            providerId = cloudProviders.stream()
                    .allMatch(curr -> curr.getProviderMetaData().getType().equalsIgnoreCase("state-aware-s3"));
        }

        final boolean supportsRedirection = cloudProviders.stream().allMatch(CloudBinaryProvider::supportsRedirection);
        final boolean cdnEntitledSupported = addonsManager.isEntitled(EntitlementType.CDN);
        return providerId && supportsRedirection && cdnEntitledSupported && AolUtils.isAol();
    }

    @Override
    public BinaryElement findBinary(ChecksumType checksumType, String checksum) {
        try {
            BinaryEntity result = binariesDao.load(checksumType, checksum);
            if (result != null) {
                return convertToBinaryElement(result);
            }
        } catch (SQLException e) {
            throw new StorageException(
                    "Storage error loading checksum ( " + checksumType.name() + "): '" + checksum + "'", e);
        }
        return null;
    }

    @Nonnull
    @Override
    public Set<BinaryElement> findBinaries(@Nullable Collection<String> checksums) {
        Set<BinaryElement> results = Sets.newHashSet();
        if (checksums == null || checksums.isEmpty()) {
            return results;
        }
        try {
            for (ChecksumType checksumType : ChecksumType.BASE_CHECKSUM_TYPES) {
                Collection<String> validChecksums = extractValid(checksumType, checksums);
                if (!validChecksums.isEmpty()) {
                    binariesDao.search(checksumType, validChecksums).stream()
                            .map(this::convertToBinaryElement)
                            .forEach(results::add);
                }
            }
        } catch (SQLException e) {
            throw new StorageException("Could not search for checksums " + checksums, e);
        }
        return results;
    }

    @Override
    public GarbageCollectorInfo runFullGarbageCollect() {
        final GarbageCollectorInfo result = new GarbageCollectorInfo(GarbageCollectorStrategy.FULL);
        notifyGCListenersOnStart(result);
        int previousCheckSumsCleaned = 0;
        Collection<BinaryEntity> binsToDelete;
        try {
            BinariesInfo countAndSize = binariesDao.getCountAndTotalSize();
            result.initialCount = countAndSize.getBinariesCount();
            result.initialSize = countAndSize.getBinariesSize();
            log.info("The initial binary count is: {}, the initial binary size is: {}",
                    result.initialCount, result.initialSize);
        } catch (SQLException e) {
            throw new StorageException("Could not find Binaries total count to delete!", e);
        }
        try {
            int iteration = 1;
            int prevDeletedBinaries = Integer.MAX_VALUE;
            while (shouldRunNextBatch((binsToDelete = binariesDao.findPotentialDeletion()).size(),
                    prevDeletedBinaries, iteration)) {
                result.candidatesForDeletion += binsToDelete.size();
                log.debug("Found {} candidates for deletion, iteration amount is {}",
                        result.candidatesForDeletion, iteration);
                List<String> deletedBinaries = deleteBinaries(result, binsToDelete);
                prevDeletedBinaries = deletedBinaries.size();
                deleteBlobInfos(deletedBinaries);
                if (result.checksumsCleaned.get() - previousCheckSumsCleaned > 0) {
                    previousCheckSumsCleaned = result.checksumsCleaned.get();
                    result.archivePathsCleaned.addAndGet(getTransactionalMe().deleteUnusedArchivePaths());
                    result.archiveNamesCleaned.addAndGet(getTransactionalMe().deleteUnusedArchiveNames());
                }
                iteration++;
            }
            log.info("Found {} candidates for deletion, iteration amount is {}",
                    result.candidatesForDeletion, iteration - 1);
        } catch (SQLException e) {
            log.error("Could not find potential binaries to delete " + e.getMessage());
            throw new StorageException("Could not find potential Binaries to delete!", e);
        }

        result.gcEndTime = System.currentTimeMillis();
        result.setStatus();
        try {
            BinariesInfo countAndSize = binariesDao.getCountAndTotalSize();
            result.printCollectionInfo(countAndSize.getBinariesSize());
        } catch (SQLException e) {
            log.error("Could not list files due to " + e.getMessage());
        }
        tryToOptimize();
        garbageCollectionDatastoreService.saveLastGCRunInfo(result);
        return result;
    }

    private boolean shouldRunNextBatch(int binsToDelete, int prevDeletedBinaries, int currIteration) {
        if (currIteration > Math.max(1, ConstantValues.gcBinariesToDeleteIterationAmount.getLong()) &&
                binsToDelete != 0 && prevDeletedBinaries != 0) {
            log.warn("The GC is stopping due to maximum iterations reached");
        }

        return binsToDelete != 0 && prevDeletedBinaries != 0 &&
                currIteration <= Math.max(1, ConstantValues.gcBinariesToDeleteIterationAmount.getLong());
    }

    @Override
    public void updateFederateRepoBinaryProvider(@Nonnull Collection<BinaryStoreFederatedMember> membersToAdd,
            @Nonnull Collection<BinaryStoreFederatedMember> membersToDelete,
                AuthenticationTokenProvider authenticationTokenProvider) {
        this.authenticationTokenProvider = authenticationTokenProvider;
        binaryProviderManager.signalMirrorChanged(membersToAdd, membersToDelete);
    }

    public AuthenticationTokenProvider getTokenProvider() {
        return this.authenticationTokenProvider;
    }

    @Autowired
    public void setArtifactoryTracer(ArtifactoryTracer artifactoryTracer) {
        this.artifactoryTracer = artifactoryTracer;
    }

    public void deleteBlobInfos(List<String> deletedBinaries) {
        if (CollectionUtils.notNullOrEmpty(deletedBinaries)) {
            deleteBlobInfosWrapper.deleteBlobInfosAsync(deletedBinaries);
        }
    }

    @Override
    public void startGarbageCollect() {
        log.info("Triggering Garbage Collection");
        int skipFullGcCount = ConstantValues.gcSkipFullGcBetweenMinorIterations.getInt();
        boolean isTwentyIteration = gcIteration.incrementAndGet() % skipFullGcCount == 0;
        try {
            TrashService trashService = ContextHelper.get().beanForType(TrashService.class);
            if (trashService.isTrashcanEnabled()) {
                if (isTwentyIteration) {
                    runAllGarbageCollectStrategies();
                } else {
                    runTrashAndBinariesGarbageCollect();
                }
            } else {
                runFullGarbageCollect();
            }
        } finally {
            if (isTwentyIteration) {
                gcIteration.set(0);
            }
        }
    }

    public void runAllGarbageCollectStrategies() {
        runTrashAndBinariesGarbageCollect();
        runFullGarbageCollect();
    }

    private void runTrashAndBinariesGarbageCollect() {
        BinariesGarbageCollectorService binariesGarbageCollectorService = ContextHelper.get()
                .beanForType(BinariesGarbageCollectorService.class);
        binariesGarbageCollectorService.startGCByStrategy(GarbageCollectorStrategy.TRASH_AND_BINARIES);
    }

    private List<String> deleteBinaries(GarbageCollectorInfo result, Collection<BinaryEntity> binsToDelete) {
        int failures = 0;
        List<String> deletedBinaries = Lists.newArrayList();
        for (BinaryEntity bd : binsToDelete) {
            log.trace("Candidate for deletion: {}", bd);
            try {
                boolean deleted =
                        dbService.invokeInTransaction("BinaryCleaner#" + bd.getSha1(), new BinaryCleaner(bd, result));
                if (deleted) {
                    deletedBinaries.add(bd.getSha2());
                }
            } catch (Exception e) {
                failures++;
                String msg = "Caught Exception, trying to clean {} : {}";
                if (failures >= ConstantValues.gcFailCountThreshold.getInt()) {
                    // We're past the allowed fail threshold, fail gc.
                    log.error("{}. Aborting Garbage Collection Run. {}, {}", msg, bd.getSha1(), e.getMessage());
                    log.debug("", e);
                    break;
                } else {
                    log.debug(msg, bd.getSha1(), e.getMessage());
                }
            }
        }
        result.failures = failures;
        return deletedBinaries;
    }

    /**
     * Deletes binary row and all dependent rows from the database
     *
     * @param sha1ToDelete Checksum to delete
     * @return True if deleted. False if not found or error
     */
    private boolean deleteEntry(String sha1ToDelete) {
        boolean hadArchiveEntries;
        try {
            hadArchiveEntries = archiveEntriesService.deleteArchiveEntries(sha1ToDelete);
        } catch (Exception e) {
            log.error("Failed to delete archive entries for " + sha1ToDelete, e);
            return false;
        }
        try {
            boolean entryDeleted = binariesDao.deleteEntry(sha1ToDelete) == 1;
            if (!entryDeleted && hadArchiveEntries) {
                log.error("Binary entry " + sha1ToDelete + " had archive entries that are deleted," +
                        " but the binary line was not deleted! Re indexing of archive needed.");
            }
            return entryDeleted;
        } catch (SQLException e) {
            log.error("Could execute delete from binary store of " + sha1ToDelete, e);
        }
        return false;
    }

    @Override
    public int deleteUnusedArchivePaths() {
        try {
            log.debug("Deleting unused archive paths");
            return archiveEntriesService.deleteUnusedPathIds();
        } catch (StorageException e) {
            log.error("Failed to delete unique paths: {}", e.getMessage());
            log.debug("Failed to delete unique paths", e);
            return 0;
        }
    }

    @Override
    public int deleteUnusedArchiveNames() {
        try {
            log.debug("Deleting unused archive names");
            return archiveEntriesService.deleteUnusedNameIds();
        } catch (StorageException e) {
            log.error("Failed to delete unique archive names: {}", e.getMessage());
            log.debug("Failed to delete unique archive paths", e);
            return 0;
        }
    }

    @Override
    public int incrementNoDeleteLock(String sha1) {
        Pair<AtomicInteger, Long> pair = deleteProtectedBinaries
                .putIfAbsent(sha1, new Pair<>(new AtomicInteger(1), System.currentTimeMillis()));
        if (pair == null) {
            return 1;
        } else {
            pair.setSecond(System.currentTimeMillis());
            return pair.getFirst().incrementAndGet();
        }
    }

    @Override
    public void decrementNoDeleteLock(String sha1) {
        AtomicInteger usageCount = deleteProtectedBinaries.get(sha1).getFirst();
        if (usageCount != null) {
            usageCount.decrementAndGet();
        }
    }

    @Override
    public boolean updateSha2ForSha1(String targetSha1, String newSha2) throws SQLException {
        return binariesDao.insertSha2(targetSha1, newSha2);
    }

    @Override
    public Collection<BinaryElement> findAllBinaries() {
        try {
            Collection<BinaryEntity> allBinaries = binariesDao.findAll();
            List<BinaryElement> result = new ArrayList<>(allBinaries.size());
            result.addAll(allBinaries.stream().map(this::convertToBinaryElement).collect(Collectors.toList()));
            return result;
        } catch (SQLException e) {
            throw new StorageException("Could not retrieve all binary entries", e);
        }
    }

    @Override
    @Nonnull
    public BinaryElement insertRecordInDb(String sha1, String sha2, String md5, long length) throws StorageException {
        BinaryEntityWithValidation dataRecord = new BinaryEntityWithValidation(sha1, sha2, md5, length);
        if (!dataRecord.isValid()) {
            throw new StorageException("Cannot insert invalid binary record: " + dataRecord);
        }
        try {
            //[sha2]: validate by sha2 as well? can find same-sha1-different-sha2 rows like this and fail uploads
            boolean binaryExists = binariesDao.exists(ChecksumType.sha1, sha1);
            if (!binaryExists) {
                createDataRecord(dataRecord, sha1);
            }
            // Always reselect from DB before returning
            BinaryEntity binary = binariesDao.load(ChecksumType.sha1, sha1);
            if (binary == null) {
                throw new StorageException("Could not find just inserted binary record: " + dataRecord);
            }
            if (StringUtils.isBlank(binary.getSha2()) && StringUtils.isNotBlank(sha2)) {
                // This could happen in case the binary was deployed before we stored sha2 in Artifactory
                binariesDao.insertSha2(binary.getSha1(), sha2);
                binary = binariesDao.load(ChecksumType.sha1, sha1);
                if (binary == null) {
                    throw new StorageException("Could not find just updated binary record: " + dataRecord);
                }
            }
            return convertToBinaryElement(binary);
        } catch (SQLException e) {
            throw new StorageException("Failed to insert new binary record: " + e.getMessage(), e);
        }
    }

    /**
     * @return Number of binaries and total size stored in the binary store
     */
    @Override
    public BinariesInfo getBinariesInfo() {
        try {
            return binariesDao.getCountAndTotalSize();
        } catch (SQLException e) {
            throw new StorageException("Could not calculate total size due to " + e.getMessage(), e);
        }
    }

    @Override
    public long getStorageSize() {
        return getTotalStorageSize();
    }

    @Override
    public void ping(boolean localOnly) {
        // Ping storage
        Map<String, String> headers = new HashMap<>();
        if (localOnly) {
            headers.put(LOCAL_ONLY.getHeaderName(), "true");
        }
        try {
            binaryProviderManager.ping(headers);
        } catch (BinaryStorageException bse) {
            log.warn("Binary provider failed ping attempt: {}", bse.getMessage());
            log.debug("", bse);
            throw bse;
        }
    }

    @Override
    public void prune(BasicStatusHolder statusHolder) {
        boolean locked = lock.tryLock();
        if (locked) {
            try {
                FilestorePruner pruner =
                        new FilestorePruner((BinaryProviderBase) binaryProviderManager.getFirstBinaryProvider(),
                                this::isActivelyUsed, BinaryServiceUtils.binariesDaoSearch(binariesDao), statusHolder);
                pruner.prune();
            } catch (Exception e) {
                log.debug("", e);
                statusHolder.error(e.getMessage(), e, log);
            } finally {
                lock.unlock();
            }
        } else {
            throw new PruneException("The prune process is already running");
        }
    }

    /**
     * @param sha1 sha1 checksum of the binary to check
     * @return True if the given binary is currently used by a reader (e.g., open stream) or writer
     */
    @Override
    public boolean isActivelyUsed(String sha1) {
        Pair<AtomicInteger, Long> pair = deleteProtectedBinaries.get(sha1);
        return pair != null && pair.getFirst().get() > 0;
    }

    private Collection<String> extractValid(ChecksumType checksumType, Collection<String> checksums) {
        return checksums.stream()
                .filter(checksumType::isValid)
                .collect(Collectors.toSet());
    }

    private InternalBinaryService getTransactionalMe() {
        return ContextHelper.get().beanForType(InternalBinaryService.class);
    }

    private BinaryElement convertToBinaryElement(BinaryEntity bd) {
        return new BinaryElementImpl(bd.getSha1(), bd.getSha2(), bd.getMd5(), bd.getLength());
    }

    private void notifyGCListenersOnStart(GarbageCollectorInfo info) {
        for (GarbageCollectorListener garbageCollectorListener : garbageCollectorListeners) {
            garbageCollectorListener.start(info);
        }
    }

    private void notifyGCListenersOnDestroy() {
        garbageCollectorListeners.forEach(GarbageCollectorListener::destroy);
    }

    private long getTotalStorageSize() {
        try {
            return binariesDao.getTotalSize();
        } catch (SQLException e) {
            throw new StorageException("Could not calculate total size due to " + e.getMessage(), e);
        }
    }

    @Override
    public List<String> getAndManageErrors() {
        List<String> errors = binaryProviderManager.getErrors();
        if (errors.size() > 0) {
            forceOptimizationOnce();
        }
        return errors;
    }

    @Override
    public void forceOptimizationOnce() {
        String FORCE_OPTIMIZATION_ONCE = "forceOptimizationOnce";
        ConflictGuard forceOptimizationOnceLock = getConflictsGuard().getLock(FORCE_OPTIMIZATION_ONCE);
        try {
            forceOptimizationOnceLock.tryToLock(0, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            log.error("Fail to get lock for force optimization");
            Thread.currentThread().interrupt();
        }
    }

    @Override
    public BinaryElement createBinaryElement(String sha1, String sha2, String md5, long length,
            Map<String, String> headers) {
        return binaryProviderManager.createBinaryElement(sha1, sha2, md5, length, headers);
    }

    @Override
    public BinaryStream createBinaryStream(InputStream in, @Nullable BinaryHeaders binaryHeaders) {
        BinaryStream binaryStream = binaryProviderManager.createBinaryStream(in, getExpectedHeaders(binaryHeaders));
        addCreateBinaryHeadersIfNeeded(binaryStream.getBinaryElement());
        return binaryStream;
    }

    /**
     * Incoming checksum info for new files can either be client checksum populated by the upload mechanism,
     * trusted if required by the uploader (i.e. ui uploads) where no checksum validation is done on the client side,
     * or empty if no such information was provided.
     * The {@param info} is used to indicate expected checksums for the Binarystore which will fail to save a binary
     * on validation error.
     *
     * {@see ConstantValues.failUploadOnChecksumValidationError} controls whether validation information is passed to
     * the Binarystore
     */
    @Nonnull
    private Map<String, String> getExpectedHeaders(@Nullable BinaryHeaders binaryHeaders) {
        Map<String, String> headers = Maps.newHashMap();
        //This flag should only be set to false if I messed up and users need to skip validations in an emergency :(

        if (binaryHeaders != null && binaryHeaders.getChecksumsInfo() != null &&
                ConstantValues.failUploadOnChecksumValidationError.getBoolean()) {
            ChecksumsInfo info = binaryHeaders.getChecksumsInfo();
            addExpectedChecksumHeader(CHECKSUM_SHA1.getHeaderName(), info.getChecksumInfo(ChecksumType.sha1), headers);
            addExpectedChecksumHeader(CHECKSUM_SHA256.getHeaderName(), info.getChecksumInfo(ChecksumType.sha256),
                    headers);
            addExpectedChecksumHeader(CHECKSUM_MD5.getHeaderName(), info.getChecksumInfo(ChecksumType.md5), headers);
        }
        if (binaryHeaders != null && binaryHeaders.getContentLength() != null) {
            String contentLength = binaryHeaders.getContentLength();
            if (!StringUtils.isEmpty(contentLength)) {
                headers.put(HttpHeaders.CONTENT_LENGTH, contentLength);
            }
        }
        if (binaryHeaders != null && binaryHeaders.getRepoKey() != null) {
            headers.put(BinaryElementHeaders.REPOSITORY_KEY.getHeaderName(), binaryHeaders.getRepoKey());

        }
        return headers;
    }

    /**
     * Adds {@param headerName} to the headers map if the {@param checksumInfo} contains data,
     * and the original (client / expected) checksum is not {@see ChecksumInfo.TRUSTED_FILE_MARKER}, which cannot be
     * used
     * for validating the actual checksum of a binary
     */
    private void addExpectedChecksumHeader(String headerName, @Nullable ChecksumInfo checksumInfo,
            Map<String, String> headers) {
        String expectedChecksum = checksumInfo == null ? null : checksumInfo.getOriginal();
        if (StringUtils.isNotBlank(expectedChecksum) && !TRUSTED_FILE_MARKER.equals(expectedChecksum)) {
            headers.put(headerName, expectedChecksum);
        }
    }

    private void createDataRecord(BinaryEntity dataRecord, String sha1) throws SQLException {
        // insert a new binary record to the db
        try {
            binariesDao.create(dataRecord);
        } catch (SQLException e) {
            if (isDuplicatedEntryException(e)) {
                log.debug("Simultaneous insert of binary {} detected, binary will be checked.", sha1, e);
                throw new BinaryInsertRetryException(convertToBinaryElement(dataRecord), e);
            } else {
                throw e;
            }
        }
    }

    public boolean isFileExist(Map<String, String> headers, String sha1) {
        BinaryElementRequestImpl request = new BinaryElementRequestImpl(sha1);
        request.getHeaders().putAll(headers);
        return binaryProviderManager.getFirstBinaryProvider().exists(request).exists();
    }

    public BinaryProvidersResponseMap isFileRestoredFromColdStorage(BinaryElementRequest request) {
        addIsBinaryRestoredHeadersIfNeeded(request);
        return binaryProviderManager.getFirstBinaryProvider().exists(request);
    }

    public void initiateRestoreFromColdStorage(BinaryElementRequest request) {
        addInitiateBinaryRestoreHeadersIfNeeded(request);
        InputStream stream = null;
        try {
            stream = getBinary(request);
        } finally {
            StreamUtils.close(stream);
        }
    }

    //taking lock if needed and only then open the transaction
    @Override
    public List<PersistedRemoteTask> getNextRemoteTaskFromDataDb(long batchSize, int maxExecutionsCount,
            long retryDelayMs,
            long maximumExecTimeMs) throws SQLException {
        DbType dbType = dbService.getDatabaseType();
        if (DbType.POSTGRESQL.equals(dbType) || DbType.MARIADB.equals(dbType) || DbType.MYSQL.equals(dbType)) {
            return getTransactionalMe().
                    getRemoteTasksBatch(batchSize, maxExecutionsCount, retryDelayMs, maximumExecTimeMs);
        } else {
            return fetchTasksWithLock(batchSize, maxExecutionsCount, retryDelayMs, maximumExecTimeMs);
        }
    }

    private List<PersistedRemoteTask> fetchTasksWithLock(long batchSize, int maxExecutionsCount, long retryDelayMs,
            long maximumExecTimeMs) throws SQLException {
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        ConflictsGuard<Object> conflictsGuard = addonsManager.addonByType(HaCommonAddon.class)
                .getConflictsGuard("federation");
        ConflictGuard lock = conflictsGuard.getLock("fetch-tasks");
        try {
            lock.tryToLock(10, TimeUnit.MINUTES);
            return getTransactionalMe().
                    getRemoteTasksBatch(batchSize, maxExecutionsCount, retryDelayMs, maximumExecTimeMs);
        } catch (InterruptedException e) {
            log.error("Timed out when fetching remote tasks", e);
            return Collections.emptyList();
        } finally {
            lock.unlock();
        }
    }

    @NotNull
    public List<PersistedRemoteTask> getRemoteTasksBatch(long batchSize, int maxExecutionsCount, long retryDelayMs,
            long maximumExecTimeMs) throws SQLException {
        List<RemoteBinaryTask> tasks;
        tasks = remoteBinaryTaskDao.getTasks(ContextHelper.get().getServerId(), batchSize,
                maxExecutionsCount, retryDelayMs, maximumExecTimeMs);

        if (tasks == null) {
            return Collections.emptyList();
        }
        return tasks.stream()
                .map(persistedRemoteTask -> new PersistedRemoteTask(
                        persistedRemoteTask.getId(),
                        persistedRemoteTask.getSha1(),
                        persistedRemoteTask.getLocalRepoKey(),
                        persistedRemoteTask.getRemoteUrlHash()
                )).collect(Collectors.toList());
    }

    private void addCreateBinaryHeadersIfNeeded(BinaryElement binaryElement) {
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        ArtifactRetentionAddon artifactRetentionAddon = addonsManager.addonByType(ArtifactRetentionAddon.class);
        addHeadersIfNeeded(artifactRetentionAddon.getCreateBinaryHeaders(), binaryElement);
    }

    private void addInitiateBinaryRestoreHeadersIfNeeded(BinaryElement binaryElement) {
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        ArtifactRetentionAddon artifactRetentionAddon = addonsManager.addonByType(ArtifactRetentionAddon.class);
        addHeadersIfNeeded(artifactRetentionAddon.getInitiateBinaryRestoreHeaders(), binaryElement);
    }

    private void addIsBinaryRestoredHeadersIfNeeded(BinaryElement binaryElement) {
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        ArtifactRetentionAddon artifactRetentionAddon = addonsManager.addonByType(ArtifactRetentionAddon.class);
        addHeadersIfNeeded(artifactRetentionAddon.getIsBinaryRestoredHeaders(), binaryElement);
    }

    private void addHeadersIfNeeded(@Nonnull Map<String, String> headers, @Nonnull BinaryElement element) {
        headers.entrySet().stream()
                .forEach(e -> element.addHeader(BinaryElementHeaders.fromHeaderName(e.getKey()), e.getValue()));
    }

    @Override
    public void onContextCreated() {
        binaryProviderManager.contextCreated();
    }

    @Override
    public void clusterTopologyChanged(List<ArtifactoryServer> activeNodes) {
        List<ClusterNode> active = activeNodes.stream()
                .filter(Objects::nonNull)
                .map(BinaryServiceUtils::clusterNodeFromArtServer)
                .collect(Collectors.toList());
        binaryProviderManager.clusterChanged(active);
    }

    @Override
    public void convert(CompoundVersionDetails source, CompoundVersionDetails target) {

    }

    public boolean executeBinaryCleaner(GCCandidate gcCandidate, GarbageCollectorInfo result) {
        return dbService.invokeInTransaction("BinaryCleaner#" + gcCandidate.getSha1(),
                new BinaryCleaner(new BinaryEntity(gcCandidate), result));
    }

    @Override
    public boolean isCloudProviderConfigured() {
        return !getCloudBinaryProviders().isEmpty();
    }

    private boolean exists(String sha1) {
        try {
            return binariesDao.exists(ChecksumType.sha1, sha1);
        } catch (SQLException sql) {
            log.debug("Failed existence check of path " + sha1, sql);
            throw new BinaryStorageException("Failed existence check of checksum " + sha1 + ": " + sql.getMessage(),
                    sql, HttpStatus.SC_INTERNAL_SERVER_ERROR);
        }
    }

    private AccessClient getAccessClient() {
        return accessService.getAccessClient();
    }

    protected void tryToOptimize() {
        ConflictGuard forceOptimizationOnceLock = getConflictsGuard().getLock("forceOptimizationOnce");
        boolean success = binaryProviderManager.optimize(forceOptimizationOnceLock.isLocked());
        if (success && forceOptimizationOnceLock.isLocked()) {
            forceOptimizationOnceLock.forceUnlock();
        }
    }

    private ConflictsGuard<String> getConflictsGuard() {
        AddonsManager addonsManager = ContextHelper.get().beanForType(AddonsManager.class);
        return addonsManager.addonByType(HaCommonAddon.class)
                .getConflictsGuard(NonStandardLockCategory.FORCE_BINARY_SHARDING_OPTIMIZATION.getName());
    }

    private int getNumOfActiveServersInCluster() {
        log.debug("Retrieving num of active servers");
        ArtifactoryServersCommonService artifactoryServersCommonService =
                ContextHelper.get().beanForType(ArtifactoryServersCommonService.class);
        return (int) artifactoryServersCommonService.getActiveMembers().stream()
                .filter(server -> server.getServerState().equals(
                        RUNNING)).count();
    }

    private Supplier<Integer> getNumOfActiveServersSupplier() {
        return Suppliers.memoizeWithExpiration(
                this::getNumOfActiveServersInCluster,
                15,
                TimeUnit.SECONDS
        );
    }

    private static class ArtifactoryLockingMapFactoryProvider implements LockingMapFactoryProvider {

        private ConflictsGuard<String> conflictsGuard;
        private ConflictsGuard<String> clusterSingletonConflictsGuard;

        @Override
        public ConflictsGuard<String> getConflictsGuard() {
            if (conflictsGuard == null) {
                conflictsGuard = ContextHelper.get().beanForType(AddonsManager.class)
                        .addonByType(HaCommonAddon.class)
                        .getConflictsGuard("binarystore");
            }
            return conflictsGuard;
        }

        @Override
        public ConflictsGuard<String> getClusterSingletonConflictsGuard() {
            if (clusterSingletonConflictsGuard == null) {
                clusterSingletonConflictsGuard = ContextHelper.get().beanForType(AddonsManager.class)
                        .addonByType(HaCommonAddon.class)
                        .getConflictsGuard(NonStandardLockCategory.CLUSTER_SINGLETON.getName());
            }
            return clusterSingletonConflictsGuard;
        }
    }

    public static class BinaryServiceInputStreamWrapper extends ReaderTrackingInputStream
            implements BinaryServiceInputStream {
        private BinaryElement bi;

        public BinaryServiceInputStreamWrapper(BinaryElement bi, UsageTracking usageTracking) {
            super(null, bi.getSha1(), usageTracking);
            this.bi = bi;
        }

        @Nonnull
        @Override
        public BinaryElement getBinaryInfo() {
            return bi;
        }
    }

    /**
     * Deletes a single binary from the database and filesystem if not in use.
     */
    class BinaryCleaner implements Callable<Boolean> {
        private final GarbageCollectorInfo result;
        private final BinaryEntity bd;

        BinaryCleaner(BinaryEntity bd, GarbageCollectorInfo result) {
            this.result = result;
            this.bd = bd;
        }

        @Override
        public Boolean call() {
            boolean binaryDeleted = false;
            String sha1 = bd.getSha1();
            deleteProtectedBinaries.putIfAbsent(sha1, new Pair<>(new AtomicInteger(0), System.currentTimeMillis()));
            Pair<AtomicInteger, Long> pair = deleteProtectedBinaries.get(sha1);
            if (pair.getFirst().compareAndSet(0, -30)) {
                log.debug("Targeting '{}' for deletion as it not seems to be used", sha1);
                try {
                    if (deleteEntry(sha1)) {
                        log.trace("Deleted {} record from binaries table", sha1);
                        result.checksumsCleaned.incrementAndGet();
                        if (binaryProviderManager.getFirstBinaryProvider().delete(new BinaryElementRequestImpl(sha1))) {
                            binaryDeleted = true;
                            log.trace("Deleted {} binary", sha1);
                            result.binariesCleaned.incrementAndGet();
                            result.totalSizeCleaned.addAndGet(bd.getLength());
                        } else {
                            log.error("Could not delete binary '{}'", sha1);
                        }
                    } else {
                        log.debug("Deleting '{}' has failed", sha1);
                    }
                } finally {
                    // remove delete protection (even if delete was not successful)
                    deleteProtectedBinaries.remove(sha1);
                    log.debug("Cleaning '{}' from ref. counter", sha1);
                }
            } else {
                Long timestamp = pair.getSecond();
                log.debug("Binary {} has {} readers with last timestamp of {}", sha1, pair.getFirst().get(), timestamp);
                long trashTime = (System.currentTimeMillis() - timestamp) / 1000;
                if (trashTime > ConstantValues.gcReadersMaxTimeSecs.getLong()) {
                    log.debug("Binary {} has reached it's max read time, removing it from ref. counter", sha1);
                    deleteProtectedBinaries.remove(sha1);
                } else {
                    log.debug("Binary {} is being read! Not deleting.", sha1);
                }
            }
            return binaryDeleted;
        }
    }
}
