/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.security.service.access;

import org.artifactory.api.security.AclService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.model.xstream.security.RepoPermissionTargetImpl;
import org.artifactory.security.AceInfo;
import org.artifactory.security.Acl;
import org.artifactory.security.BuildPermissionTarget;
import org.artifactory.security.MutableAceInfo;
import org.artifactory.security.PermissionTarget;
import org.artifactory.security.PrincipalPermission;
import org.artifactory.security.ReleaseBundlePermissionTarget;
import org.artifactory.security.RepoAcl;
import org.artifactory.security.RepoPermissionTarget;
import org.artifactory.security.access.AccessService;
import org.artifactory.storage.db.security.service.AclCacheLoader;
import org.artifactory.storage.db.security.service.AllAclCacheLoader;
import org.artifactory.storage.db.security.service.VersionCache;
import org.artifactory.test.ArtifactoryHomeStub;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.permission.PermissionsClient;
import org.jfrog.access.common.ServiceId;
import org.jfrog.common.ClockUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.ReentrantLock;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

/**
 * @author Noam Shemesh
 */
public class AccessAclStoreServiceTest {
    private AclStoreServiceImpl aclStoreServiceImpl;

    @Mock
    private AccessService accessService;
    @Mock
    private PermissionsClient permissionsClient;
    @Mock
    private AclService aclService;
    @Mock
    VersionCache repoAclsCache;
    @Mock
    private VersionCache<AclCacheLoader.AclCacheItem<ReleaseBundlePermissionTarget>> releaseBundlesAclsCache;
    @Mock
    private VersionCache<AclCacheLoader.AclCacheItem<BuildPermissionTarget>> buildAclsCache;
    @Mock
    VersionCache<AllAclCacheLoader.AllAclCacheItem> allAclsCache;
    @Mock
    AclCacheLoader.AclCacheItem<RepoPermissionTarget> aclCacheItem;

    @BeforeMethod
    public void beforeTest() {
        MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(new ArtifactoryHomeStub());
        aclStoreServiceImpl = new AclStoreServiceImpl();
        aclStoreServiceImpl.setAccessService(accessService);
        aclStoreServiceImpl.setAclService(aclService);
        aclStoreServiceImpl.init();

        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.aclMinimumWaitRefreshDataSeconds.getPropertyName(), "2");
        AccessClient accessClient = mock(AccessClient.class);
        when(accessClient.permissions()).thenReturn(permissionsClient);
        when(accessService.getAccessClient()).thenReturn(accessClient);
        when(accessService.getArtifactoryServiceId()).thenReturn(ServiceId.fromFormattedName("jfrt@service"));

        when(repoAclsCache.promoteVersion()).thenReturn(1);
        when(releaseBundlesAclsCache.promoteVersion()).thenReturn(1);
        when(buildAclsCache.promoteVersion()).thenReturn(1);
        when(allAclsCache.promoteVersion()).thenReturn(1);
        ReflectionTestUtils.setField(aclStoreServiceImpl, "repoAclsCache", repoAclsCache);
        ReflectionTestUtils.setField(aclStoreServiceImpl, "releaseBundlesAclsCache", releaseBundlesAclsCache);
        ReflectionTestUtils.setField(aclStoreServiceImpl, "buildsAclsCache", buildAclsCache);
        ReflectionTestUtils.setField(aclStoreServiceImpl, "allAclsCache", allAclsCache);
        Map<String, Map<String, Set<PrincipalPermission<RepoPermissionTarget>>>> groupResultMap = new HashMap<>();
        when(aclCacheItem.getGroupResultMap()).thenReturn(groupResultMap);
        when(repoAclsCache.get()).thenReturn(aclCacheItem);
        ReentrantLock lock = new ReentrantLock();
        ReflectionTestUtils.setField(aclStoreServiceImpl, "lock", lock);
    }

    @Test
    public void testRefreshingCacheMultiThreaded() throws InterruptedException {
        List<Thread> threadList = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            when(permissionsClient.lastUpdated()).thenReturn(ClockUtils.epochMillis() + 60000);
            threadList.add(new Thread(() -> aclStoreServiceImpl.getAclCache()));
        }
        threadList.forEach(Thread::run);
        for (Thread thread : threadList) {
            thread.join();
        }
        verify(repoAclsCache, times(1)).promoteVersion();
    }

    @Test(enabled = false)
    public void testRefreshingCache() throws InterruptedException {
        when(permissionsClient.lastUpdated()).thenReturn(ClockUtils.epochMillis() + 60000);
        aclStoreServiceImpl.getAclCache();
        aclStoreServiceImpl.getAclCache();
        Thread.sleep(2500);
        aclStoreServiceImpl.getAclCache();

        verify(repoAclsCache, times(2)).promoteVersion();
    }

    @Test(enabled = false)
    public void testSkipRefreshingCacheWhenAccessIsNotRefreshed() throws InterruptedException {
        when(permissionsClient.lastUpdated()).thenReturn(ClockUtils.epochMillis() - 20000);
        aclStoreServiceImpl.getAclCache();
        Thread.sleep(2500);
        aclStoreServiceImpl.getAclCache();
        verify(repoAclsCache, times(1)).promoteVersion();
    }

    @Test
    public void testUpdateAclIsSkippedForInternal() {
        when(permissionsClient.lastUpdated()).thenReturn(ClockUtils.epochMillis());
        Acl<? extends PermissionTarget> acl = new MockAcl();
        when(aclService.isInternalAcl(acl)).thenReturn(false);
        aclStoreServiceImpl.updateAcl(acl);
        verify(permissionsClient).replacePermission(any());
        clearInvocations(permissionsClient);

        when(aclService.isInternalAcl(acl)).thenReturn(true);
        aclStoreServiceImpl.updateAcl(acl);
        verifyNoMoreInteractions(permissionsClient);
    }

    public static class MockAcl implements RepoAcl {

        @Override
        public RepoPermissionTarget getPermissionTarget() {
            return new RepoPermissionTargetImpl("permissionTarget");
        }

        @Override
        public Set<AceInfo> getAces() {
            return Collections.emptySet();
        }

        @Override
        public Set<MutableAceInfo> getMutableAces() {
            return Collections.emptySet();
        }

        @Override
        public String getUpdatedBy() {
            return null;
        }

        @Override
        public long getLastUpdated() {
            return 0;
        }

        @Nullable
        @Override
        public String getAccessIdentifier() {
            return "identifier";
        }
    }
}